# -*- coding: utf-8 -*-
import calendar
import datetime
import math

import arrow
import pandas as pd
import tzlocal

from zvtm.contract import IntervalLevel

CHINA_TZ = 'Asia/Shanghai'

TIME_FORMAT_ISO8601 = "YYYY-MM-DDTHH:mm:ss.SSS"

TIME_FORMAT_MON = 'YYYY-MM'

TIME_FORMAT_DAY = 'YYYY-MM-DD'

TIME_FORMAT_DAY1 = 'YYYYMMDD'

TIME_FORMAT_MINUTE = 'YYYYMMDDHHmm'

TIME_FORMAT_MINUTE1 = 'HH:mm'

TIME_FORMAT_MINUTE2 = "YYYY-MM-DD HH:mm:ss"


# ms(int) or second(float) or str
def to_pd_timestamp(the_time) -> pd.Timestamp:
    if the_time is None:
        return None
    if type(the_time) == int:
        return pd.Timestamp.fromtimestamp(the_time / 1000)

    if type(the_time) == float:
        return pd.Timestamp.fromtimestamp(the_time)

    return pd.Timestamp(the_time)


def to_timestamp(the_time):
    return int(to_pd_timestamp(the_time).tz_localize(tzlocal.get_localzone()).timestamp() * 1000)


def now_timestamp():
    return int(pd.Timestamp.utcnow().timestamp() * 1000)


def now_pd_timestamp() -> pd.Timestamp:
    return pd.Timestamp.now()


def today() -> pd.Timestamp:
    return pd.Timestamp.today()


def current_date() -> pd.Timestamp:
    return to_pd_timestamp(today().date())


def to_time_str(the_time, fmt=TIME_FORMAT_DAY):
    try:
        return arrow.get(to_pd_timestamp(the_time)).format(fmt)
    except Exception as e:
        return the_time


def now_time_str(fmt=TIME_FORMAT_DAY):
    return to_time_str(the_time=now_pd_timestamp(), fmt=fmt)


def next_date(the_time, days=1):
    return to_pd_timestamp(the_time) + datetime.timedelta(days=days)


def pre_month(t=now_pd_timestamp()):
    t = to_pd_timestamp(t)
    t = t.replace(day=1)
    if t.month > 1:
        year = t.year
        month = t.month - 1
    else:
        year = t.year - 1
        month = 12
    last_valid_date = t.replace(year=year, month=month)
    return last_valid_date


def pre_month_start_date(t=current_date()):
    return month_start_date(pre_month(t))


def pre_month_end_date(t=current_date()):
    return month_end_date(pre_month(t))


def month_start_date(the_date):
    the_date = to_pd_timestamp(the_date)
    return the_date.replace(day=1)


def month_end_date(the_date):
    the_date = to_pd_timestamp(the_date)

    _, day = calendar.monthrange(the_date.year, the_date.month)
    return the_date.replace(day=day)


def month_start_end_ranges(start_date, end_date):
    days = pd.date_range(start=start_date, end=end_date, freq='M')
    return [(month_start_date(d), month_end_date(d)) for d in days]


def is_same_date(one, two):
    return to_pd_timestamp(one).date() == to_pd_timestamp(two).date()


def is_same_time(one, two):
    return to_timestamp(one) == to_timestamp(two)


def get_year_quarter(time):
    time = to_pd_timestamp(time)
    return time.year, ((time.month - 1) // 3) + 1


def day_offset_today(offset=0):
    return now_pd_timestamp() + datetime.timedelta(days=offset)


def get_year_quarters(start, end=pd.Timestamp.now()):
    start_year_quarter = get_year_quarter(start)
    current_year_quarter = get_year_quarter(end)
    if current_year_quarter[0] == start_year_quarter[0]:
        return [(current_year_quarter[0], x) for x in range(start_year_quarter[1], current_year_quarter[1] + 1)]
    elif current_year_quarter[0] - start_year_quarter[0] == 1:
        return [(start_year_quarter[0], x) for x in range(start_year_quarter[1], 5)] + \
               [(current_year_quarter[0], x) for x in range(1, current_year_quarter[1] + 1)]
    elif current_year_quarter[0] - start_year_quarter[0] > 1:
        return [(start_year_quarter[0], x) for x in range(start_year_quarter[1], 5)] + \
               [(x, y) for x in range(start_year_quarter[0] + 1, current_year_quarter[0]) for y in range(1, 5)] + \
               [(current_year_quarter[0], x) for x in range(1, current_year_quarter[1] + 1)]
    else:
        raise Exception("wrong start time:{}".format(start))


def date_and_time(the_date, the_time):
    time_str = '{}T{}:00.000'.format(to_time_str(the_date), the_time)

    return to_pd_timestamp(time_str)


def next_timestamp(current_timestamp: pd.Timestamp, level: IntervalLevel) -> pd.Timestamp:
    current_timestamp = to_pd_timestamp(current_timestamp)
    return current_timestamp + pd.Timedelta(seconds=level.to_second())


def evaluate_size_from_timestamp(start_timestamp,
                                 level: IntervalLevel,
                                 one_day_trading_minutes,
                                 end_timestamp: pd.Timestamp = None):
    """
    given from timestamp,level,one_day_trading_minutes,this func evaluate size of kdata to current.
    it maybe a little bigger than the real size for fetching all the kdata.

    :param start_timestamp:
    :type start_timestamp: pd.Timestamp
    :param level:
    :type level: IntervalLevel
    :param one_day_trading_minutes:
    :type one_day_trading_minutes: int
    """
    if not end_timestamp:
        end_timestamp = pd.Timestamp.now()
    else:
        end_timestamp = to_pd_timestamp(end_timestamp)

    time_delta = end_timestamp - to_pd_timestamp(start_timestamp)

    one_day_trading_seconds = one_day_trading_minutes * 60

    if level == IntervalLevel.LEVEL_1DAY:
        return time_delta.days + 1

    if level == IntervalLevel.LEVEL_1WEEK:
        return int(math.ceil(time_delta.days / 7)) + 1

    if level == IntervalLevel.LEVEL_1MON:
        return int(math.ceil(time_delta.days / 30)) + 1

    if time_delta.days > 0:
        seconds = (time_delta.days + 1) * one_day_trading_seconds
        return int(math.ceil(seconds / level.to_second())) + 1
    else:
        seconds = time_delta.total_seconds()
        return min(int(math.ceil(seconds / level.to_second())) + 1,
                   one_day_trading_seconds / level.to_second() + 1)


def is_finished_kdata_timestamp(timestamp, level: IntervalLevel):
    timestamp = to_pd_timestamp(timestamp)
    if level.floor_timestamp(timestamp) == timestamp:
        return True
    return False


def is_in_same_interval(t1: pd.Timestamp, t2: pd.Timestamp, level: IntervalLevel):
    t1 = to_pd_timestamp(t1)
    t2 = to_pd_timestamp(t2)
    if level == IntervalLevel.LEVEL_1WEEK:
        return t1.week == t2.week
    if level == IntervalLevel.LEVEL_1MON:
        return t1.month == t2.month

    return level.floor_timestamp(t1) == level.floor_timestamp(t2)


def split_time_interval(start, end, method=None, interval=30, freq='D'):
    start = to_pd_timestamp(start)
    end = to_pd_timestamp(end)
    if not method:
        while start < end:
            interval_end = min(next_date(start, interval), end)
            yield pd.date_range(start=start, end=interval_end, freq=freq)
            start = next_date(interval_end, 1)

    if method == 'month':
        while start <= end:
            _, day = calendar.monthrange(start.year, start.month)

            interval_end = min(to_pd_timestamp(f'{start.year}-{start.month}-{day}'), end)
            yield pd.date_range(start=start, end=interval_end, freq=freq)
            start = next_date(interval_end, 1)


def count_interval(start_date, end_date):
    start_date = to_pd_timestamp(start_date)
    end_date = to_pd_timestamp(end_date)
    delta = end_date - start_date
    return delta.days


if __name__ == '__main__':
    print(date_and_time('2019-10-01', '10:00'))
# the __all__ is generated
__all__ = ['to_pd_timestamp', 'to_timestamp', 'now_timestamp', 'now_pd_timestamp', 'today', 'to_time_str',
           'now_time_str', 'next_date', 'pre_month_start_date', 'month_start_date', 'month_end_date',
           'month_start_end_ranges', 'is_same_date', 'is_same_time', 'get_year_quarter', 'day_offset_today',
           'get_year_quarters', 'date_and_time', 'next_timestamp', 'evaluate_size_from_timestamp',
           'is_finished_kdata_timestamp', 'is_in_same_interval', 'split_time_interval']



# -*- coding: utf-8 -*-
# this file is generated by gen_kdata_schema function, dont't change it
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule etf_1d_kdata
from .etf_1d_kdata import *
from .etf_1d_kdata import __all__ as _etf_1d_kdata_all
__all__ += _etf_1d_kdata_all


# -*- coding: utf-8 -*-

from sqlalchemy.orm import declarative_base

from zvtm.contract.register import register_schema
from zvtm.contract.schema import ActorEntity

ActorMetaBase = declarative_base()


# 参与者
class ActorMeta(ActorMetaBase, ActorEntity):
    __tablename__ = 'actor_meta'


register_schema(providers=['em'], db_name='actor_meta', schema_base=ActorMetaBase)
# the __all__ is generated
__all__ = ['ActorMeta']


# -*- coding: utf-8 -*-
from zvtm.contract import IntervalLevel
from zvtm.domain import ReportPeriod


def to_jq_trading_level(trading_level: IntervalLevel):
    if trading_level < IntervalLevel.LEVEL_1HOUR:
        return trading_level.value

    if trading_level == IntervalLevel.LEVEL_1HOUR:
        return '60m'
    if trading_level == IntervalLevel.LEVEL_4HOUR:
        return '240m'
    if trading_level == IntervalLevel.LEVEL_1DAY:
        return '1d'
    if trading_level == IntervalLevel.LEVEL_1WEEK:
        return '1w'
    if trading_level == IntervalLevel.LEVEL_1MON:
        return '1M'


def to_jq_entity_id(security_item):
    if security_item.entity_type == 'stock' or security_item.entity_type == 'index':
        if security_item.exchange == 'sh':
            return '{}.XSHG'.format(security_item.code)
        if security_item.exchange == 'sz':
            return '{}.XSHE'.format(security_item.code)


def to_entity_id(jq_code: str, entity_type):
    try:
        code, exchange = jq_code.split('.')
        if exchange == 'XSHG':
            exchange = 'sh'
        elif exchange == 'XSHE':
            exchange = 'sz'
    except:
        code = jq_code
        exchange = 'sz'

    return f'{entity_type}_{exchange}_{code}'


def jq_to_report_period(jq_report_type):
    if jq_report_type == '第一季度':
        return ReportPeriod.season1.value
    if jq_report_type == '第二季度':
        return ReportPeriod.season2.value
    if jq_report_type == '第三季度':
        return ReportPeriod.season3.value
    if jq_report_type == '第四季度':
        return ReportPeriod.season4.value
    if jq_report_type == '半年度':
        return ReportPeriod.half_year.value
    if jq_report_type == '年度':
        return ReportPeriod.year.value
    assert False


# the __all__ is generated
__all__ = ['to_jq_trading_level', 'to_jq_entity_id', 'to_entity_id', 'jq_to_report_period']


# -*- coding: utf-8 -*-#
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule algorithm
from .algorithm import *
from .algorithm import __all__ as _algorithm_all
__all__ += _algorithm_all

# import all from submodule ma
from .ma import *
from .ma import __all__ as _ma_all
__all__ += _ma_all

# import all from submodule macd
from .macd import *
from .macd import __all__ as _macd_all
__all__ += _macd_all

# import all from submodule zen
from .zen import *
from .zen import __all__ as _zen_all
__all__ += _zen_all

# import all from submodule technical_factor
from .technical_factor import *
from .technical_factor import __all__ as _technical_factor_all
__all__ += _technical_factor_all

# import all from submodule fundamental
from .fundamental import *
from .fundamental import __all__ as _fundamental_all
__all__ += _fundamental_all

# import all from submodule target_selector
from .target_selector import *
from .target_selector import __all__ as _target_selector_all
__all__ += _target_selector_all


# -*- coding: utf-8 -*-
from zvtm.api.utils import to_report_period_type
from zvtm.domain import IncomeStatement
from zvtm.recorders.eastmoney.finance.base_china_stock_finance_recorder import BaseChinaStockFinanceRecorder
from zvtm.utils.time_utils import to_pd_timestamp
from zvtm.utils.utils import add_func_to_value, first_item_to_float

income_statement_map = {
    # 营业总收入
    #
    # 营业收入
    "operating_income": "Operatereve",
    # 营业总成本
    "total_operating_costs": "Totaloperateexp",
    # 营业成本
    "operating_costs": "Operateexp",
    # 研发费用
    "rd_costs": "Rdexp",
    # 提取保险合同准备金净额
    "net_change_in_insurance_contract_reserves": "Netcontactreserve",
    # 营业税金及附加
    "business_taxes_and_surcharges": "Operatetax",
    # 销售费用
    "sales_costs": "Saleexp",
    # 管理费用
    "managing_costs": "Manageexp",
    # 财务费用
    "financing_costs": "Financeexp",
    # 资产减值损失
    "assets_devaluation": "Assetdevalueloss",
    # 其他经营收益
    #
    # 加: 投资收益
    "investment_income": "Investincome",
    # 其中: 对联营企业和合营企业的投资收益
    "investment_income_from_related_enterprise": "Investjointincome",
    # 营业利润
    "operating_profit": "Operateprofit",
    # 加: 营业外收入
    "non_operating_income": "Nonoperatereve",
    # 减: 营业外支出
    "non_operating_costs": "Nonoperateexp",
    # 其中: 非流动资产处置净损失
    "loss_on_disposal_non_current_asset": "Nonlassetnetloss",

    # 利润总额
    "total_profits": "Sumprofit",
    # 减: 所得税费用
    "tax_expense": "Incometax",
    # 净利润
    "net_profit": "Netprofit",
    # 其中: 归属于母公司股东的净利润
    "net_profit_as_parent": "Parentnetprofit",
    # 少数股东损益
    "net_profit_as_minority_interest": "Minorityincome",
    # 扣除非经常性损益后的净利润
    "deducted_net_profit": "Kcfjcxsyjlr",
    # 每股收益
    # 基本每股收益
    "eps": "Basiceps",
    # 稀释每股收益
    "diluted_eps": "Dilutedeps",
    # 其他综合收益
    "other_comprehensive_income": "Othercincome",
    # 归属于母公司股东的其他综合收益
    "other_comprehensive_income_as_parent": "Parentothercincome",
    # 归属于少数股东的其他综合收益
    "other_comprehensive_income_as_minority_interest": "Minorityothercincome",
    # 综合收益总额
    "total_comprehensive_income": "Sumcincome",
    # 归属于母公司所有者的综合收益总额
    "total_comprehensive_income_as_parent": "Parentcincome",
    # 归属于少数股东的综合收益总额
    "total_comprehensive_income_as_minority_interest": "Minoritycincome",

    # 银行相关
    # 利息净收入
    "fi_net_interest_income": "Intnreve",
    # 其中:利息收入
    "fi_interest_income": "Intreve",
    # 利息支出
    "fi_interest_expenses": "Intexp",
    # 手续费及佣金净收入
    "fi_net_incomes_from_fees_and_commissions": "Commnreve",
    # 其中:手续费及佣金收入
    "fi_incomes_from_fees_and_commissions": "Commreve",
    # 手续费及佣金支出
    "fi_expenses_for_fees_and_commissions": "Commexp",
    # 公允价值变动收益
    "fi_income_from_fair_value_change": "Fvalueincome",
    # 汇兑收益
    "fi_income_from_exchange": "Exchangeincome",
    # 其他业务收入
    "fi_other_income": "Otherreve",
    # 业务及管理费
    "fi_operate_and_manage_expenses": "Operatemanageexp",

    # 保险相关
    # 已赚保费
    "fi_net_income_from_premium": "Premiumearned",
    # 其中:保险业务收入
    "fi_income_from_premium": "Insurreve",
    # 分保费收入
    "fi_income_from_reinsurance_premium": "Rireve",
    # 减:分出保费
    "fi_reinsurance_premium": "Ripremium",
    # 提取未到期责任准备金
    "fi_undue_duty_reserve": "Unduereserve",
    # 银行业务利息净收入
    "fi_net_income_from_bank_interest": "Bankintnreve",
    # 其中:银行业务利息收入
    "fi_income_from_bank_interest": "Bankintreve",
    # 银行业务利息支出
    "fi_expenses_for_bank_interest": "Bankintexp",
    # 非保险业务手续费及佣金净收入
    "fi_net_incomes_from_fees_and_commissions_of_non_insurance": "Ninsurcommnreve",
    # 非保险业务手续费及佣金收入
    "fi_incomes_from_fees_and_commissions_of_non_insurance": "Ninsurcommreve",
    # 非保险业务手续费及佣金支出
    "fi_expenses_for_fees_and_commissions_of_non_insurance": "Ninsurcommexp",
    # 退保金
    "fi_insurance_surrender_costs": "Surrenderpremium",
    # 赔付支出
    "fi_insurance_claims_expenses": "Indemnityexp",
    # 减:摊回赔付支出
    "fi_amortized_insurance_claims_expenses": "Amortiseindemnityexp",
    # 提取保险责任准备金
    "fi_insurance_duty_reserve": "Dutyreserve",
    # 减:摊回保险责任准备金
    "fi_amortized_insurance_duty_reserve": "Amortisedutyreserve",
    # 保单红利支出
    "fi_dividend_expenses_to_insured": "Policydiviexp",
    # 分保费用
    "fi_reinsurance_expenses": "Riexp",
    # 减:摊回分保费用
    "fi_amortized_reinsurance_expenses": "Amortiseriexp",
    # 其他业务成本
    "fi_other_op_expenses": "Otherexp",

    # 券商相关
    # 手续费及佣金净收入
    #
    # 其中:代理买卖证券业务净收入
    "fi_net_incomes_from_trading_agent": "Agenttradesecurity",
    # 证券承销业务净收入
    "fi_net_incomes_from_underwriting": "Securityuw",
    # 受托客户资产管理业务净收入
    "fi_net_incomes_from_customer_asset_management": "Clientassetmanage",
    # 手续费及佣金净收入其他项目
    "fi_fees_from_other": "Commnreveother",
    # 公允价值变动收益
    #
    # 其中:可供出售金融资产公允价值变动损益
    "fi_income_from_fair_value_change_of_fi_salable": "Fvalueosalable",
}

add_func_to_value(income_statement_map, first_item_to_float)
income_statement_map["report_period"] = ("ReportDate", to_report_period_type)
income_statement_map["report_date"] = ("ReportDate", to_pd_timestamp)


class ChinaStockIncomeStatementRecorder(BaseChinaStockFinanceRecorder):
    data_schema = IncomeStatement

    url = 'https://emh5.eastmoney.com/api/CaiWuFenXi/GetLiRunBiaoList'
    finance_report_type = 'LiRunBiaoList'

    data_type = 2

    def get_data_map(self):
        return income_statement_map


if __name__ == '__main__':
    # init_log('income_statement.log')
    recorder = ChinaStockIncomeStatementRecorder(codes=['002572'])
    recorder.run()
# the __all__ is generated
__all__ = ['ChinaStockIncomeStatementRecorder']


# -*- coding: utf-8 -*-
from typing import List

import pandas as pd

from zvtm.api.utils import to_report_period_type, value_to_pct
from zvtm.contract import ActorType
from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import TimestampsDataRecorder
from zvtm.domain import Stock
from zvtm.domain.actor.stock_actor import StockActorSummary
from zvtm.recorders.em.em_api import get_ii_holder_report_dates, actor_type_to_org_type, get_ii_summary
from zvtm.utils import to_pd_timestamp, to_time_str


# [{'CHANGE_RATIO': -1.045966694333,
#   'IS_COMPLETE': '1',
#   'ORG_TYPE': '07',
#   'REPORT_DATE': '2021-03-31 00:00:00',
#   'SECUCODE': '000338.SZ',
#   'SECURITY_CODE': '000338',
#   'TOTAL_FREE_SHARES': 2598718411,
#   'TOTAL_MARKET_CAP': 49999342227.64,
#   'TOTAL_ORG_NUM': 5,
#   'TOTAL_SHARES_RATIO': 29.51742666}]

class EMStockActorSummaryRecorder(TimestampsDataRecorder):
    entity_provider = 'joinquant'
    entity_schema = Stock

    provider = 'em'
    data_schema = StockActorSummary

    def init_timestamps(self, entity_item) -> List[pd.Timestamp]:
        result = get_ii_holder_report_dates(code=entity_item.code)
        if result:
            return [to_pd_timestamp(item['REPORT_DATE']) for item in result]

    def record(self, entity, start, end, size, timestamps):
        for timestamp in timestamps:
            the_date = to_time_str(timestamp)
            self.logger.info(f'to {entity.code} {the_date}')
            for actor_type in ActorType:
                if actor_type == ActorType.private_equity or actor_type == ActorType.individual:
                    continue
                result = get_ii_summary(code=entity.code, report_date=the_date,
                                        org_type=actor_type_to_org_type(actor_type))
                if result:
                    summary_list = [{'id': f'{entity.entity_id}_{the_date}_{actor_type.value}',
                                     'entity_id': entity.entity_id,
                                     'timestamp': timestamp,
                                     'code': entity.code,
                                     'name': entity.name,

                                     'actor_type': actor_type.value,
                                     'actor_count': item['TOTAL_ORG_NUM'],

                                     'report_date': timestamp,
                                     'report_period': to_report_period_type(timestamp),

                                     'change_ratio': value_to_pct(item['CHANGE_RATIO'], default=1),
                                     'is_complete': item['IS_COMPLETE'],
                                     'holding_numbers': item['TOTAL_FREE_SHARES'],
                                     'holding_ratio': value_to_pct(item['TOTAL_SHARES_RATIO'], default=0),
                                     'holding_values': item['TOTAL_MARKET_CAP']
                                     } for item in result]
                    df = pd.DataFrame.from_records(summary_list)
                    df_to_db(data_schema=self.data_schema, df=df, provider=self.provider,
                             force_update=True, drop_duplicates=True)


if __name__ == '__main__':
    EMStockActorSummaryRecorder(codes=['000338']).run()
# the __all__ is generated
__all__ = ['EMStockActorSummaryRecorder']


from jqdatapy.api import run_query

from zvtm.contract.recorder import TimeSeriesDataRecorder
from zvtm.domain import Index, CrossMarketSummary
from zvtm.utils.time_utils import to_time_str
from zvtm.utils.utils import multiple_number


class CrossMarketSummaryRecorder(TimeSeriesDataRecorder):
    entity_provider = 'joinquant'
    entity_schema = Index

    provider = 'joinquant'
    data_schema = CrossMarketSummary

    def __init__(self,
                 force_update=False, sleeping_time=5, real_time=False,
                 fix_duplicate_way='add') -> None:

        # 聚宽编码
        # 市场通编码	市场通名称
        # 310001	沪股通
        # 310002	深股通
        # 310003	港股通（沪）
        # 310004	港股通（深）

        codes = ['310001', '310002', '310003', '310004']
        super().__init__(force_update, sleeping_time, ['cn'], None, codes=codes, day_data=True, real_time=real_time,
                         fix_duplicate_way=fix_duplicate_way)

    def init_entities(self):
        super().init_entities()

    def record(self, entity, start, end, size, timestamps):
        df = run_query(table='finance.STK_ML_QUOTA', conditions=f'link_id#=#{entity.code}&day#>=#{to_time_str(start)}')
        print(df)

        json_results = []

        for item in df.to_dict(orient='records'):
            result = {
                'provider': self.provider,
                'timestamp': item['day'],
                'name': entity.name,
                'buy_amount': multiple_number(item['buy_amount'], 100000000),
                'buy_volume': item['buy_volume'],
                'sell_amount': multiple_number(item['sell_amount'], 100000000),
                'sell_volume': item['sell_volume'],
                'quota_daily': multiple_number(item['quota_daily'], 100000000),
                'quota_daily_balance': multiple_number(item['quota_daily_balance'], 100000000)
            }

            json_results.append(result)

        if len(json_results) < 100:
            self.one_shot = True

        return json_results

    def get_data_map(self):
        return None


if __name__ == '__main__':
    CrossMarketSummaryRecorder().run()
# the __all__ is generated
__all__ = ['CrossMarketSummaryRecorder']


# -*- coding: utf-8 -*-

from sqlalchemy import Column, String
from sqlalchemy.orm import declarative_base

from zvtm.contract import Portfolio, PortfolioStock
from zvtm.contract.register import register_schema, register_entity

BlockMetaBase = declarative_base()


# 板块
@register_entity(entity_type='block')
class Block(BlockMetaBase, Portfolio):
    __tablename__ = 'block'

    # 板块类型，行业(industry),概念(concept)
    category = Column(String(length=64))


class BlockStock(BlockMetaBase, PortfolioStock):
    __tablename__ = 'block_stock'


register_schema(providers=['eastmoney', 'sina'], db_name='block_meta',
                schema_base=BlockMetaBase)
# the __all__ is generated
__all__ = ['Block', 'BlockStock']


# -*- coding: utf-8 -*-
import os
from pathlib import Path

# zvt home dir
ZVT_HOME = os.environ.get('ZVTM_HOME')
if not ZVT_HOME:
    ZVT_HOME = os.path.abspath(os.path.join(Path.home(), 'zvtm-home'))

# data for testing
ZVT_TEST_HOME = os.path.abspath(os.path.join(Path.home(), 'zvtm-test-home'))
ZVT_TEST_ZIP_DATA_PATH = os.path.join(ZVT_TEST_HOME, 'data.zip')
ZVT_TEST_DATA_PATH = os.path.join(ZVT_TEST_HOME, 'data')

DATA_SAMPLE_ZIP_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), 'samples', 'data.zip'))

# ****** setting for stocks ****** #
# 覆盖维度 银行/保险/企业/券商 创业板 中小板 主板
SAMPLE_STOCK_CODES = ['000001', '000783', '000778', '603220', '601318', '000338', '002572', '300027']

# 沪深300，证券，中证500，上证50，创业板，军工,传媒,资源
SAMPLE_ETF_CODES = ['510300', '512880', '510500', '510050', '159915', '512660', '512980', '510410']



# -*- coding: utf-8 -*-
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule informer
from .informer import *
from .informer import __all__ as _informer_all
__all__ += _informer_all


# -*- coding: utf-8 -*-
from sqlalchemy import Column, String, Integer
from sqlalchemy.orm import declarative_base

from zvtm.contract import Portfolio, PortfolioStockHistory
from zvtm.contract.register import register_entity, register_schema
from zvtm.utils import now_pd_timestamp

FundMetaBase = declarative_base()


# 个股
@register_entity(entity_type='fund')
class Fund(FundMetaBase, Portfolio):
    __tablename__ = 'fund'
    # 基金管理人
    advisor = Column(String(length=100))
    # 基金托管人
    trustee = Column(String(length=100))

    # 编码	基金运作方式
    # 401001	开放式基金
    # 401002	封闭式基金
    # 401003	QDII
    # 401004	FOF
    # 401005	ETF
    # 401006	LOF
    # 基金运作方式编码
    operate_mode_id = Column(Integer)
    # 基金运作方式
    operate_mode = Column(String(length=32))

    # 编码	基金类别
    # 402001	股票型
    # 402002	货币型
    # 402003	债券型
    # 402004	混合型
    # 402005	基金型
    # 402006	贵金属
    # 402007	封闭式
    # 投资标的类型编码
    underlying_asset_type_id = Column(Integer)
    # 投资标的类型
    underlying_asset_type = Column(String(length=32))

    @classmethod
    def get_stocks(cls, code=None, codes=None, ids=None, timestamp=now_pd_timestamp(), provider=None):
        from zvtm.api.portfolio import get_fund_stocks
        return get_fund_stocks(code=code, codes=codes, ids=ids, timestamp=timestamp, provider=provider)


class FundStock(FundMetaBase, PortfolioStockHistory):
    __tablename__ = 'fund_stock'


register_schema(providers=['joinquant'], db_name='fund_meta', schema_base=FundMetaBase)
# the __all__ is generated
__all__ = ['Fund', 'FundStock']


# -*- coding: utf-8 -*-
from zvtm.contract import IntervalLevel
from zvtm.factors.target_selector import TargetSelector
from zvtm.factors.ma.ma_factor import CrossMaFactor
from zvtm.factors import BullFactor
from zvtm.trader.trader import StockTrader


class MyMaTrader(StockTrader):
    def init_selectors(self, entity_ids, entity_schema, exchanges, codes, start_timestamp, end_timestamp,
                       adjust_type=None):
        myselector = TargetSelector(entity_ids=entity_ids, entity_schema=entity_schema, exchanges=exchanges,
                                    codes=codes, start_timestamp=start_timestamp, end_timestamp=end_timestamp,
                                    provider='joinquant')

        myselector.add_filter_factor(
            CrossMaFactor(entity_ids=entity_ids, entity_schema=entity_schema, exchanges=exchanges,
                          codes=codes, start_timestamp=start_timestamp, end_timestamp=end_timestamp,
                          windows=[5, 10], need_persist=False, adjust_type=adjust_type))

        self.selectors.append(myselector)


class MyBullTrader(StockTrader):
    def init_selectors(self, entity_ids, entity_schema, exchanges, codes, start_timestamp, end_timestamp,
                       adjust_type=None):
        myselector = TargetSelector(entity_ids=entity_ids, entity_schema=entity_schema, exchanges=exchanges,
                                    codes=codes, start_timestamp=start_timestamp, end_timestamp=end_timestamp,
                                    provider='joinquant')

        myselector.add_filter_factor(
            BullFactor(entity_ids=entity_ids, entity_schema=entity_schema, exchanges=exchanges,
                       codes=codes, start_timestamp=start_timestamp, end_timestamp=end_timestamp,
                       adjust_type=adjust_type))

        self.selectors.append(myselector)


if __name__ == '__main__':
    # single stock with cross ma factor
    MyMaTrader(codes=['000338'], level=IntervalLevel.LEVEL_1DAY, start_timestamp='2018-01-01',
               end_timestamp='2019-06-30', trader_name='000338_ma_trader').run()

    # single stock with bull factor
    # MyBullTrader(codes=['000338'], level=IntervalLevel.LEVEL_1DAY, start_timestamp='2018-01-01',
    #              end_timestamp='2019-06-30', trader_name='000338_bull_trader').run()

    #  multiple stocks with cross ma factor
    # MyMaTrader(codes=SAMPLE_STOCK_CODES, level=IntervalLevel.LEVEL_1DAY, start_timestamp='2018-01-01',
    #            end_timestamp='2019-06-30', trader_name='sample_stocks_ma_trader').run()

    # multiple stocks with bull factor
    # MyBullTrader(codes=SAMPLE_STOCK_CODES, level=IntervalLevel.LEVEL_1DAY, start_timestamp='2018-01-01',
    #              end_timestamp='2019-06-30', trader_name='sample_stocks_bull_trader').run()



# -*- coding: utf-8 -*-
from zvtm.api.utils import to_report_period_type
from zvtm.domain.misc.holder import TopTenHolder
from zvtm.recorders.eastmoney.common import EastmoneyTimestampsDataRecorder, get_fc
from zvtm.utils.time_utils import to_time_str, to_pd_timestamp
from zvtm.utils.utils import to_float


class TopTenHolderRecorder(EastmoneyTimestampsDataRecorder):
    provider = 'eastmoney'
    data_schema = TopTenHolder

    url = 'https://emh5.eastmoney.com/api/GuBenGuDong/GetShiDaGuDong'
    path_fields = ['ShiDaGuDongList']

    timestamps_fetching_url = 'https://emh5.eastmoney.com/api/GuBenGuDong/GetFirstRequest2Data'
    timestamp_list_path_fields = ['SDGDBGQ', 'ShiDaGuDongBaoGaoQiList']
    timestamp_path_fields = ['BaoGaoQi']

    def get_data_map(self):
        return {
            "report_period": ("timestamp", to_report_period_type),
            "report_date": ("timestamp", to_pd_timestamp),
            # 股东代码
            "holder_code": ("GuDongDaiMa", str),
            # 股东名称
            "holder_name": ("GuDongMingCheng", str),
            # 持股数
            "shareholding_numbers": ("ChiGuShu", to_float),
            # 持股比例
            "shareholding_ratio": ("ChiGuBiLi", to_float),
            # 变动
            "change": ("ZengJian", to_float),
            # 变动比例
            "change_ratio": ("BianDongBiLi", to_float),
        }

    def generate_request_param(self, security_item, start, end, size, timestamp):
        return {"color": "w",
                "fc": get_fc(security_item),
                "BaoGaoQi": to_time_str(timestamp)
                }

    def generate_domain_id(self, entity, original_data):
        the_name = original_data.get("GuDongMingCheng")
        timestamp = original_data[self.get_original_time_field()]
        the_id = "{}_{}_{}".format(entity.id, timestamp, the_name)
        return the_id


if __name__ == '__main__':
    # init_log('top_ten_holder.log')

    TopTenHolderRecorder(codes=['002572']).run()
# the __all__ is generated
__all__ = ['TopTenHolderRecorder']


# -*- coding: utf-8 -*-
from typing import List

import pandas as pd
import requests

from zvtm.api.utils import china_stock_code_to_id, value_to_pct, value_multiply
from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import Recorder, TimestampsDataRecorder
from zvtm.domain import Index, IndexCategory, IndexStock
from zvtm.recorders.consts import DEFAULT_HEADER
from zvtm.utils.time_utils import to_pd_timestamp, to_time_str, TIME_FORMAT_MON, pre_month


# 深证指数，国证指数
class ExchangeCNIndexRecorder(Recorder):
    provider = 'exchange'
    data_schema = Index
    original_page_url = 'http://www.cnindex.com.cn/zh_indices/sese/index.html?act_menu=1&index_type=-1'
    url = 'http://www.cnindex.net.cn/index/indexList?channelCode={}&rows=1000&pageNum=1'

    # 中证指数 抓取 风格指数 行业指数 规模指数 基金指数
    cni_category_map_url = {
        IndexCategory.style: url.format('202'),
        IndexCategory.industry: url.format('201'),
        IndexCategory.scope: url.format('200'),
        IndexCategory.fund: url.format('207'),
    }

    # 深证指数 只取规模指数
    sz_category_map_url = {
        IndexCategory.scope: url.format('100'),
    }

    def run(self):
        self.record_index('sz')
        self.record_index('cni')

    def get_resp_data(self, resp: requests.Response):
        resp.raise_for_status()
        return resp.json()['data']

    def record_index(self, index_type):
        if index_type == 'cni':
            category_map_url = self.cni_category_map_url
        elif index_type == 'sz':
            category_map_url = self.sz_category_map_url
        else:
            self.logger.error(f'not support index_type: {index_type}')
            assert False

        requests_session = requests.Session()

        for category, url in category_map_url.items():
            resp = requests_session.get(url, headers=DEFAULT_HEADER)

            results = self.get_resp_data(resp)['rows']
            # e.g
            # amount: 277743699997.9
            # closeingPoint: 6104.7592
            # docchannel: 1039
            # freeMarketValue: 10794695531696.15
            # id: 142
            # indexcode: "399370"
            # indexename: "CNI Growth"
            # indexfullcname: "国证1000成长指数"
            # indexfullename: "CNI 1000 Growth Index"
            # indexname: "国证成长"
            # indexsource: "1"
            # indextype: "202"
            # pb: 5.34
            # peDynamic: 29.8607
            # peStatic: 33.4933
            # percent: 0.0022
            # prefixmonth: null
            # realtimemarket: "1"
            # remark: ""
            # sampleshowdate: null
            # samplesize: 332
            # showcnindex: "1"
            # totalMarketValue: 23113641352198.32
            the_list = []

            self.logger.info(f'category: {category} ')
            self.logger.info(f'results: {results} ')
            for i, result in enumerate(results):
                self.logger.info(f'to {i}/{len(results)}')
                code = result['indexcode']
                info_resp = requests_session.get(f'http://www.cnindex.net.cn/index-intro?indexcode={code}')
                # fbrq: "2010-01-04"
                # jd: 1000
                # jr: "2002-12-31"
                # jsfs: "自由流通市值"
                # jsjj: "国证成长由国证1000指数样本股中成长风格突出的股票组成，为投资者提供更丰富的指数化投资工具。"
                # qzsx: null
                # typl: 2
                # xyfw: "沪深A股"
                # xygz: "在国证1000指数样本股中，选取主营业务收入增长率、净利润增长率和净资产收益率综合排名前332只"
                index_info = self.get_resp_data(info_resp)
                name = result['indexname']
                entity_id = f'index_sz_{code}'
                index_item = {
                    'id': entity_id,
                    'entity_id': entity_id,
                    'timestamp': to_pd_timestamp(index_info['jr']),
                    'entity_type': 'index',
                    'exchange': 'sz',
                    'code': code,
                    'name': name,
                    'category': category.value,
                    'list_date': to_pd_timestamp(index_info['fbrq']),
                    'base_point': index_info['jd'],
                    'publisher': 'cnindex'
                }
                self.logger.info(index_item)
                the_list.append(index_item)
                self.sleep(2)
            if the_list:
                df = pd.DataFrame.from_records(the_list)
                df_to_db(data_schema=self.data_schema, df=df, provider=self.provider,
                         force_update=True)
            self.logger.info(f"finish record {index_type} index:{category.value}")


class ExchangeCNIndexStockRecorder(TimestampsDataRecorder):
    entity_provider = 'exchange'
    entity_schema = Index

    provider = 'exchange'
    data_schema = IndexStock

    original_page_url = 'http://www.cnindex.com.cn/module/index-detail.html?act_menu=1&indexCode=399001'
    url = 'http://www.cnindex.net.cn/sample-detail/detail?indexcode={}&dateStr={}&pageNum=1&rows=5000'

    def __init__(self,
                 force_update=False,
                 sleeping_time=5,
                 exchanges=None,
                 entity_ids=None,
                 codes=None,
                 day_data=False,
                 entity_filters=None,
                 ignore_failed=True,
                 real_time=False,
                 fix_duplicate_way='add',
                 start_timestamp=None,
                 end_timestamp=None,
                 record_history=False) -> None:
        super().__init__(force_update, sleeping_time, exchanges, entity_ids, codes, day_data, entity_filters,
                         ignore_failed, real_time, fix_duplicate_way, start_timestamp, end_timestamp)
        self.record_history = record_history

    def init_timestamps(self, entity_item) -> List[pd.Timestamp]:
        last_valid_date = pre_month()
        if self.record_history:
            # 每个月记录一次
            return [to_pd_timestamp(item) for item in
                    pd.date_range(entity_item.list_date, last_valid_date, freq='M')]
        else:
            return [last_valid_date]

    def get_resp_data(self, resp: requests.Response):
        resp.raise_for_status()
        return resp.json()['data']

    def record(self, entity, start, end, size, timestamps):
        for timestamp in timestamps:
            data_str = to_time_str(timestamp, TIME_FORMAT_MON)
            resp = requests.get(self.url.format(entity.code, data_str), headers=DEFAULT_HEADER)
            results = self.get_resp_data(resp)['rows']
            if not results:
                self.logger.warning(f'no data for timestamp: {data_str}')
                self.sleep(3)
                continue
            the_list = []
            for result in results:
                # date: 1614268800000
                # dateStr: "2021-02-26"
                # freeMarketValue: 10610.8
                # indexcode: "399370"
                # market: null
                # seccode: "600519"
                # secname: "贵州茅台"
                # totalMarketValue: 26666.32
                # trade: "主要消费"
                # weight: 10.01
                stock_code = result['indexcode']
                stock_name = result['secname']
                stock_id = china_stock_code_to_id(stock_code)

                the_list.append({
                    'id': '{}_{}_{}'.format(entity.id, result['dateStr'], stock_id),
                    'entity_id': entity.id,
                    'entity_type': entity.entity_type,
                    'exchange': entity.exchange,
                    'code': entity.code,
                    'name': entity.name,
                    'timestamp': to_pd_timestamp(result['dateStr']),
                    'stock_id': stock_id,
                    'stock_code': stock_code,
                    'stock_name': stock_name,
                    'proportion': value_to_pct(result['weight'], 0),
                    'market_cap': value_multiply(result['freeMarketValue'], 100000000, 0)
                })
            if the_list:
                df = pd.DataFrame.from_records(the_list)
                df_to_db(data_schema=self.data_schema, df=df, provider=self.provider, force_update=True)

            self.logger.info(f'finish recording index:{entity.id}, {entity.name}, {data_str}')

            self.sleep(3)


if __name__ == '__main__':
    # init_log('china_stock_category.log')
    # ExchangeCNIndexRecorder().run()

    recorder = ExchangeCNIndexStockRecorder(codes=['399370'])
    recorder.run()
# the __all__ is generated
__all__ = ['ExchangeCNIndexRecorder', 'ExchangeCNIndexStockRecorder']


# -*- coding: utf-8 -*-
import logging
import os
import platform
from typing import List, Union, Type

import pandas as pd
import pymysql
pymysql.install_as_MySQLdb()
from sqlalchemy import create_engine
from sqlalchemy import func, exists, and_
from sqlalchemy.engine import Engine
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm import Query
from sqlalchemy.orm import sessionmaker, Session
from sqlalchemy_utils import database_exists,create_database
from zvtm import zvt_env
from zvtm.contract import IntervalLevel, TradableEntity
from zvtm.contract import Mixin
from zvtm.contract import zvt_context
from zvtm.utils.pd_utils import pd_is_not_null, index_df
from zvtm.utils.time_utils import to_pd_timestamp

from zvtm import zvt_config

logger = logging.getLogger(__name__)


def get_db_name(data_schema: DeclarativeMeta) -> str:
    """
    get db name of the domain schema

    :param data_schema:
    :type data_schema:
    :return:
    :rtype:
    """
    for db_name, base in zvt_context.dbname_map_base.items():
        if issubclass(data_schema, base):
            return db_name


def get_db_engine(provider: str,
                  db_name: str = None,
                  data_schema: object = None,
                  data_path: str = zvt_env['data_path']) -> Engine:
    """
    get db engine of the (provider,db_name) or (provider,data_schema)


    :param data_path:
    :param provider:
    :type provider:
    :param db_name:
    :type db_name:
    :param data_schema:
    :type data_schema:
    :return:
    :rtype:
    """
    if data_schema:
        db_name = get_db_name(data_schema=data_schema)

    db_path = os.path.join(data_path, '{}_{}.db?check_same_thread=False'.format(provider, db_name))

    engine_key = '{}_{}'.format(provider, db_name)
    db_engine = zvt_context.db_engine_map.get(engine_key)
    #engine_url = 'mysql+mysqldb://root:root@localhost:3306/'
    engine_url = zvt_config['engine_url']
    if not db_engine:
        db_engine = create_engine(engine_url+engine_key+'?charset=utf8',pool_size=100,pool_recycle=3600,pool_pre_ping=True,echo=False) #jqdata?charset=utf8
        if not database_exists(db_engine.url):
            create_database(engine_url+engine_key+'?charset=utf8')
            db_engine = create_engine(engine_url + engine_key + '?charset=utf8')
        zvt_context.db_engine_map[engine_key] = db_engine
    return db_engine


def get_schemas(provider: str) -> List[DeclarativeMeta]:
    """
    get domain schemas supported by the provider

    :param provider:
    :type provider:
    :return:
    :rtype:
    """
    schemas = []
    for provider1, dbs in zvt_context.provider_map_dbnames.items():
        if provider == provider1:
            for dbname in dbs:
                schemas1 = zvt_context.dbname_map_schemas.get(dbname)
                if schemas1:
                    schemas += schemas1
    return schemas


def get_db_session(provider: str,
                   db_name: str = None,
                   data_schema: object = None,
                   force_new: bool = False) -> Session:
    """
    get db session of the (provider,db_name) or (provider,data_schema)

    :param provider:
    :type provider:
    :param db_name:
    :type db_name:
    :param data_schema:
    :type data_schema:
    :param force_new:
    :type force_new:

    :return:
    :rtype:
    """
    if data_schema:
        db_name = get_db_name(data_schema=data_schema)

    session_key = '{}_{}'.format(provider, db_name)

    if force_new:
        return get_db_session_factory(provider, db_name, data_schema)()

    session = zvt_context.sessions.get(session_key)
    if not session:
        session = get_db_session_factory(provider, db_name, data_schema)()
        zvt_context.sessions[session_key] = session
    return session


def get_db_session_factory(provider: str,
                           db_name: str = None,
                           data_schema: object = None):
    """
    get db session factory of the (provider,db_name) or (provider,data_schema)

    :param provider:
    :type provider:
    :param db_name:
    :type db_name:
    :param data_schema:
    :type data_schema:
    :return:
    :rtype:
    """
    if data_schema:
        db_name = get_db_name(data_schema=data_schema)

    session_key = '{}_{}'.format(provider, db_name)
    session = zvt_context.db_session_map.get(session_key)
    if not session:
        session = sessionmaker()
        zvt_context.db_session_map[session_key] = session
    return session


def domain_name_to_table_name(domain_name: str) -> str:
    parts = []
    part = ''
    for c in domain_name:
        if c.isupper() or c.isdigit():
            if part:
                parts.append(part)
            part = c.lower()
        else:
            part = part + c

    parts.append(part)

    if len(parts) > 1:
        return '_'.join(parts)
    elif parts:
        return parts[0]


def table_name_to_domain_name(table_name: str) -> DeclarativeMeta:
    """
    the rules for table_name -> domain_class

    :param table_name:
    :type table_name:
    :return:
    :rtype:
    """
    parts = table_name.split('_')
    domain_name = ''
    for part in parts:
        domain_name = domain_name + part.capitalize()
    return domain_name


def get_entity_schema(entity_type: str) -> object:
    """
    get entity schema from name

    :param entity_type:
    :type entity_type:
    :return:
    :rtype:
    """
    return zvt_context.tradable_schema_map[entity_type]


def get_schema_by_name(name: str) -> DeclarativeMeta:
    """
    get domain schema by the name

    :param name:
    :type name:
    :return:
    :rtype:
    """
    for schema in zvt_context.schemas:
        if schema.__name__ == name:
            return schema


def get_schema_columns(schema: DeclarativeMeta) -> object:
    """
    get all columns of the domain schema

    :param schema:
    :type schema:
    :return:
    :rtype:
    """
    return schema.__table__.columns.keys()


def common_filter(query: Query,
                  data_schema,
                  start_timestamp=None,
                  end_timestamp=None,
                  filters=None,
                  order=None,
                  limit=None,
                  time_field='timestamp'):
    assert data_schema is not None
    time_col = eval('data_schema.{}'.format(time_field))

    if start_timestamp:
        query = query.filter(time_col >= to_pd_timestamp(start_timestamp))
    if end_timestamp:
        query = query.filter(time_col <= to_pd_timestamp(end_timestamp))

    if filters:
        for filter in filters:
            query = query.filter(filter)
    if order is not None:
        query = query.order_by(order)
    else:
        query = query.order_by(time_col.asc())
    if limit:
        query = query.limit(limit)

    return query


def del_data(data_schema: Type[Mixin], filters: List = None, provider=None):
    if not provider:
        provider = data_schema.providers[0]

    session = get_db_session(provider=provider, data_schema=data_schema)
    query = session.query(data_schema)
    if filters:
        for f in filters:
            query = query.filter(f)
    query.delete()
    session.commit()


def get_one(data_schema, id: str, provider: str = None, session: Session = None):
    if 'providers' not in data_schema.__dict__:
        logger.error("no provider registered for: {}", data_schema)
    if not provider:
        provider = data_schema.providers[0]

    if not session:
        session = get_db_session(provider=provider, data_schema=data_schema)

    try:
        ids = session.query(data_schema).get(id)
        return ids
    except:
        session.close()
        #db_engine.dispose()
        session = get_db_session(provider=provider, data_schema=data_schema)
    finally:
        pass

    return session.query(data_schema).get(id)


def get_data(data_schema: Mixin,
             ids: List[str] = None,
             entity_ids: List[str] = None,
             entity_id: str = None,
             codes: List[str] = None,
             code: str = None,
             level: Union[IntervalLevel, str] = None,
             provider: str = None,
             columns: List = None,
             col_label: dict = None,
             return_type: str = 'df',
             start_timestamp: Union[pd.Timestamp, str] = None,
             end_timestamp: Union[pd.Timestamp, str] = None,
             filters: List = None,
             session: Session = None,
             order=None,
             limit: int = None,
             index: Union[str, list] = None,
             time_field: str = 'timestamp'):
    if 'providers' not in data_schema.__dict__:
        logger.error("no provider registered for: {}", data_schema)
    if not provider:
        provider = data_schema.providers[0]

    if not session:
        session = get_db_session(provider=provider, data_schema=data_schema)

    time_col = eval('data_schema.{}'.format(time_field))

    if columns:
        # support str
        if type(columns[0]) == str:
            columns_ = []
            for col in columns:
                assert isinstance(col, str)
                columns_.append(eval('data_schema.{}'.format(col)))
            columns = columns_

        # make sure get timestamp
        if time_col not in columns:
            columns.append(time_col)

        if col_label:
            columns_ = []
            for col in columns:
                if col.name in col_label:
                    columns_.append(col.label(col_label.get(col.name)))
                else:
                    columns_.append(col)
            columns = columns_

        query = session.query(*columns)
    else:
        query = session.query(data_schema)

    if entity_id:
        query = query.filter(data_schema.entity_id == entity_id)
    if entity_ids:
        query = query.filter(data_schema.entity_id.in_(entity_ids))
    if code:
        query = query.filter(data_schema.code == code)
    if codes:
        query = query.filter(data_schema.code.in_(codes))
    if ids:
        query = query.filter(data_schema.id.in_(ids))

    # we always store different level in different schema,the level param is not useful now
    if level:
        try:
            # some schema has no level,just ignore it
            data_schema.level
            if type(level) == IntervalLevel:
                level = level.value
            query = query.filter(data_schema.level == level)
        except Exception as e:
            pass

    query = common_filter(query, data_schema=data_schema, start_timestamp=start_timestamp,
                          end_timestamp=end_timestamp, filters=filters, order=order, limit=limit,
                          time_field=time_field)

    if return_type == 'df':
        df = pd.read_sql(query.statement, query.session.bind)
        if pd_is_not_null(df):
            if index:
                df = index_df(df, index=index, time_field=time_field)
        return df
    elif return_type == 'domain':
        return query.all()
    elif return_type == 'dict':
        return [item.__dict__ for item in query.all()]


def data_exist(session, schema, id):
    return session.query(exists().where(and_(schema.id == id))).scalar()


def get_data_count(data_schema, filters=None, session=None):
    query = session.query(data_schema)
    if filters:
        for filter in filters:
            query = query.filter(filter)

    count_q = query.statement.with_only_columns([func.count(data_schema.id)]).order_by(None)
    count = session.execute(count_q).scalar()
    return count


def get_group(provider, data_schema, column, group_func=func.count, session=None):
    if not session:
        session = get_db_session(provider=provider, data_schema=data_schema)
    if group_func:
        query = session.query(column, group_func(column)).group_by(column)
    else:
        query = session.query(column).group_by(column)
    df = pd.read_sql(query.statement, query.session.bind)
    return df


def decode_entity_id(entity_id: str):
    result = entity_id.split('_')
    entity_type = result[0]
    exchange = result[1]
    code = ''.join(result[2:])
    return entity_type, exchange, code


def get_entity_type(entity_id: str):
    entity_type, _, _ = decode_entity_id(entity_id)
    return entity_type


def get_entity_exchange(entity_id: str):
    _, exchange, _ = decode_entity_id(entity_id)
    return exchange


def get_entity_code(entity_id: str):
    _, _, code = decode_entity_id(entity_id)
    return code


def df_to_db(df: pd.DataFrame,
             data_schema: DeclarativeMeta,
             provider: str,
             force_update: bool = False,
             sub_size: int = 5000,
             drop_duplicates: bool = True) -> object:
    """
    FIXME:improve
    store the df to db

    :param df:
    :param data_schema:
    :param provider:
    :param force_update:
    :param sub_size:
    :param drop_duplicates:
    :return:
    """
    if not pd_is_not_null(df):
        return 0

    if drop_duplicates and df.duplicated(subset='id').any():
        logger.warning(f'remove duplicated:{df[df.duplicated()]}')
        df = df.drop_duplicates(subset='id', keep='last')

    db_engine = get_db_engine(provider, data_schema=data_schema)

    schema_cols = get_schema_columns(data_schema)
    cols = set(df.columns.tolist()) & set(schema_cols)

    if not cols:
        print('wrong cols')
        return 0

    df = df[cols]

    size = len(df)

    if platform.system() == "Windows":
        sub_size = 900

    if size >= sub_size:
        step_size = int(size / sub_size)
        if size % sub_size:
            step_size = step_size + 1
    else:
        step_size = 1

    saved = 0

    for step in range(step_size):
        df_current = df.iloc[sub_size * step:sub_size * (step + 1)]
        if force_update:
            session = get_db_session(provider=provider, data_schema=data_schema)
            ids = df_current["id"].tolist()
            if len(ids) == 1:
                sql = f'delete from `{data_schema.__tablename__}` where id = "{ids[0]}"'
            else:
                sql = f'delete from `{data_schema.__tablename__}` where id in {tuple(ids)}'

            session.execute(sql)
            try:
                session.commit()
            except:
                session.rollback()
                raise
            finally:
                session.close()
                db_engine.dispose()


        else:
            current = get_data(data_schema=data_schema, columns=[data_schema.id], provider=provider,
                               ids=df_current['id'].tolist())
            if pd_is_not_null(current):
                df_current = df_current[~df_current['id'].isin(current['id'])]

        if pd_is_not_null(df_current):
            saved = saved + len(df_current)
            df_current.to_sql(data_schema.__tablename__, db_engine, index=False, if_exists='append')

    return saved


def get_entities(
        entity_schema: TradableEntity = None,
        entity_type: str = None,
        exchanges: List[str] = None,
        ids: List[str] = None,
        entity_ids: List[str] = None,
        entity_id: str = None,
        codes: List[str] = None,
        code: str = None,
        provider: str = None,
        columns: List = None,
        col_label: dict = None,
        return_type: str = 'df',
        start_timestamp: Union[pd.Timestamp, str] = None,
        end_timestamp: Union[pd.Timestamp, str] = None,
        filters: List = None,
        session: Session = None,
        order=None,
        limit: int = None,
        index: Union[str, list] = 'code') -> object:
    if not entity_schema:
        entity_schema = zvt_context.tradable_schema_map[entity_type]

    if not provider:
        provider = entity_schema.providers[0]

    if not order:
        order = entity_schema.code.asc()

    if exchanges:
        if filters:
            filters.append(entity_schema.exchange.in_(exchanges))
        else:
            filters = [entity_schema.exchange.in_(exchanges)]

    return get_data(data_schema=entity_schema, ids=ids, entity_ids=entity_ids, entity_id=entity_id, codes=codes,
                    code=code, level=None, provider=provider, columns=columns, col_label=col_label,
                    return_type=return_type, start_timestamp=start_timestamp, end_timestamp=end_timestamp,
                    filters=filters, session=session, order=order, limit=limit, index=index)


def get_entity_ids(entity_type='stock', entity_schema: TradableEntity = None, exchanges=None, codes=None, provider=None,
                   filters=None):
    df = get_entities(entity_type=entity_type, entity_schema=entity_schema, exchanges=exchanges, codes=codes,
                      provider=provider, filters=filters)
    if pd_is_not_null(df):
        return df['entity_id'].to_list()
    return None


# the __all__ is generated
__all__ = ['get_db_name', 'get_db_engine', 'get_schemas', 'get_db_session', 'get_db_session_factory', 'domain_name_to_table_name', 'table_name_to_domain_name', 'get_entity_schema', 'get_schema_by_name', 'get_schema_columns', 'common_filter', 'del_data', 'get_one', 'get_data', 'data_exist', 'get_data_count', 'get_group', 'decode_entity_id', 'get_entity_type', 'get_entity_exchange', 'get_entity_code', 'df_to_db', 'get_entities', 'get_entity_ids']


# -*- coding: utf-8 -*-
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule stock_actor
from .stock_actor import *
from .stock_actor import __all__ as _stock_actor_all
__all__ += _stock_actor_all

# import all from submodule actor_meta
from .actor_meta import *
from .actor_meta import __all__ as _actor_meta_all
__all__ += _actor_meta_all


# -*- coding: utf-8 -*-
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule em_stock_top_ten_free_recorder
from .em_stock_top_ten_free_recorder import *
from .em_stock_top_ten_free_recorder import __all__ as _em_stock_top_ten_free_recorder_all
__all__ += _em_stock_top_ten_free_recorder_all

# import all from submodule em_stock_actor_summary_recorder
from .em_stock_actor_summary_recorder import *
from .em_stock_actor_summary_recorder import __all__ as _em_stock_actor_summary_recorder_all
__all__ += _em_stock_actor_summary_recorder_all

# import all from submodule em_stock_top_ten_recorder
from .em_stock_top_ten_recorder import *
from .em_stock_top_ten_recorder import __all__ as _em_stock_top_ten_recorder_all
__all__ += _em_stock_top_ten_recorder_all

# import all from submodule em_stock_ii_recorder
from .em_stock_ii_recorder import *
from .em_stock_ii_recorder import __all__ as _em_stock_ii_recorder_all
__all__ += _em_stock_ii_recorder_all


# -*- coding: utf-8 -*-
from typing import List, Union

import pandas as pd


def pd_is_not_null(df: Union[pd.DataFrame, pd.Series]):
    return df is not None and not df.empty


def index_df(df, index='timestamp', inplace=True, drop=False, time_field='timestamp'):
    if time_field:
        df[time_field] = pd.to_datetime(df[time_field])

    if inplace:
        df.set_index(index, drop=drop, inplace=inplace)
    else:
        df = df.set_index(index, drop=drop, inplace=inplace)

    if type(index) == str:
        df = df.sort_index()
    elif type(index) == list:
        df.index.names = index
        level = list(range(len(index)))
        df = df.sort_index(level=level)
    return df


def normal_index_df(df, category_field='entity_id', time_filed='timestamp', drop=True):
    index = [category_field, time_filed]
    if is_normal_df(df):
        return df

    return index_df(df=df, index=index, drop=drop, time_field='timestamp')


def is_normal_df(df, category_field='entity_id', time_filed='timestamp'):
    if pd_is_not_null(df):
        names = df.index.names

        if len(names) == 2 and names[0] == category_field and names[1] == time_filed:
            return True

    return False


def df_subset(df, columns=None):
    if columns:
        return df.loc[:, columns]
    return df


def fill_with_same_index(df_list: List[pd.DataFrame]):
    idx = None
    for df in df_list:
        if idx is None:
            idx = df.index
        else:
            idx = idx.append(df.index).drop_duplicates()
    idx = idx.sort_values()

    result = []
    for df in df_list:
        # print(df[df.index.duplicated()])
        added_index = idx.difference(df.index.drop_duplicates())
        added_df = pd.DataFrame(index=added_index, columns=df.columns)

        # df1 = df.reindex(idx)
        df1 = df.append(added_df)
        df1 = df1.sort_index()
        result.append(df1)
    return result
# the __all__ is generated
__all__ = ['pd_is_not_null', 'index_df', 'normal_index_df', 'is_normal_df', 'df_subset', 'fill_with_same_index']


# -*- coding: utf-8 -*-
from sqlalchemy.orm import declarative_base

from zvtm.contract import Mixin
from zvtm.contract.register import register_schema

TradeDayBase = declarative_base()


class StockTradeDay(TradeDayBase, Mixin):
    __tablename__ = 'stock_trade_day'


register_schema(providers=['joinquant'], db_name='trade_day', schema_base=TradeDayBase)

# the __all__ is generated
__all__ = ['StockTradeDay']


# -*- coding: utf-8 -*-
from sqlalchemy.ext.declarative import declarative_base

from zvtm.contract.register import register_schema
from zvtm.factors.zen.domain.common import ZenFactorCommon

Stock1dZenFactorBase = declarative_base()


class Stock1dZenFactor(Stock1dZenFactorBase, ZenFactorCommon):
    __tablename__ = 'stock_1d_zen_factor'


register_schema(providers=['zvt'], db_name='stock_1d_zen_factor', schema_base=Stock1dZenFactorBase, entity_type='stock')
# the __all__ is generated
__all__ = ['Stock1dZenFactor']


# -*- coding: utf-8 -*-
from zvtm.domain import TopTenTradableHolder
from zvtm.recorders.eastmoney.holder.eastmoney_top_ten_holder_recorder import TopTenHolderRecorder


class TopTenTradableHolderRecorder(TopTenHolderRecorder):
    provider = 'eastmoney'
    data_schema = TopTenTradableHolder

    url = 'https://emh5.eastmoney.com/api/GuBenGuDong/GetShiDaLiuTongGuDong'
    path_fields = ['ShiDaLiuTongGuDongList']
    timestamps_fetching_url = 'https://emh5.eastmoney.com/api/GuBenGuDong/GetFirstRequest2Data'
    timestamp_list_path_fields = ['SDLTGDBGQ', 'ShiDaLiuTongGuDongBaoGaoQiList']
    timestamp_path_fields = ['BaoGaoQi']


if __name__ == '__main__':
    # init_log('top_ten_tradable_holder.log')

    TopTenTradableHolderRecorder(codes=['002572']).run()
# the __all__ is generated
__all__ = ['TopTenTradableHolderRecorder']


# -*- coding: utf-8 -*-
from sqlalchemy import Column, String, Float
from sqlalchemy.orm import declarative_base

from zvtm.contract import Mixin
from zvtm.contract.register import register_schema

MoneyFlowBase = declarative_base()


# 板块资金流向

class BlockMoneyFlow(MoneyFlowBase, Mixin):
    __tablename__ = 'block_money_flow'

    code = Column(String(length=32))
    name = Column(String(length=32))

    # 收盘价
    close = Column(Float)
    change_pct = Column(Float)
    turnover_rate = Column(Float)

    # 净流入
    net_inflows = Column(Float)
    # 净流入率
    net_inflow_rate = Column(Float)

    # 主力=超大单+大单
    net_main_inflows = Column(Float)
    net_main_inflow_rate = Column(Float)
    # 超大单
    net_huge_inflows = Column(Float)
    net_huge_inflow_rate = Column(Float)
    # 大单
    net_big_inflows = Column(Float)
    net_big_inflow_rate = Column(Float)

    # 中单
    net_medium_inflows = Column(Float)
    net_medium_inflow_rate = Column(Float)
    # 小单
    net_small_inflows = Column(Float)
    net_small_inflow_rate = Column(Float)


class StockMoneyFlow(MoneyFlowBase, Mixin):
    __tablename__ = 'stock_money_flow'

    code = Column(String(length=32))
    name = Column(String(length=32))

    # 收盘价
    close = Column(Float)
    change_pct = Column(Float)
    turnover_rate = Column(Float)

    # 净流入
    net_inflows = Column(Float)
    # 净流入率
    net_inflow_rate = Column(Float)

    # 主力=超大单+大单
    net_main_inflows = Column(Float)
    net_main_inflow_rate = Column(Float)
    # 超大单
    net_huge_inflows = Column(Float)
    net_huge_inflow_rate = Column(Float)
    # 大单
    net_big_inflows = Column(Float)
    net_big_inflow_rate = Column(Float)

    # 中单
    net_medium_inflows = Column(Float)
    net_medium_inflow_rate = Column(Float)
    # 小单
    net_small_inflows = Column(Float)
    net_small_inflow_rate = Column(Float)


class IndexMoneyFlow(MoneyFlowBase, Mixin):
    __tablename__ = 'index_money_flow'

    code = Column(String(length=32))
    name = Column(String(length=32))

    # 净流入
    net_inflows = Column(Float)
    # 净流入率
    net_inflow_rate = Column(Float)

    # 主力=超大单+大单
    net_main_inflows = Column(Float)
    net_main_inflow_rate = Column(Float)
    # 超大单
    net_huge_inflows = Column(Float)
    net_huge_inflow_rate = Column(Float)
    # 大单
    net_big_inflows = Column(Float)
    net_big_inflow_rate = Column(Float)

    # 中单
    net_medium_inflows = Column(Float)
    net_medium_inflow_rate = Column(Float)
    # 小单
    net_small_inflows = Column(Float)
    net_small_inflow_rate = Column(Float)


register_schema(providers=['joinquant', 'sina'], db_name='money_flow', schema_base=MoneyFlowBase, entity_type='stock')

# the __all__ is generated
__all__ = ['BlockMoneyFlow', 'StockMoneyFlow', 'IndexMoneyFlow']


# -*- coding: utf-8 -*-
import requests

from zvtm.api import get_recent_report_date
from zvtm.contract.recorder import Recorder
from zvtm.domain.actor.actor_meta import ActorMeta
from zvtm.utils import to_pd_timestamp


class EastmoneyActorRecorder(Recorder):
    name = "eastmoney_actor_recorder"
    provider = 'eastmoney'
    data_schema = ActorMeta

    url = "https://datacenter.eastmoney.com/securities/api/data/v1/get?reportName=RPT_FREEHOLDERS_BASIC_INFO&columns=HOLDER_NAME,END_DATE,HOLDER_NEW,HOLDER_NUM,HOLDER_CODE&quoteColumns=&filter=(END_DATE='{}')&pageNumber={}&pageSize={}&sortTypes=-1,-1&sortColumns=HOLDER_NUM,HOLDER_NAME&source=SECURITIES&client=SW"

    start = '2016-03-31'

    def get_data(self, end_date, pn, ps):
        resp = requests.get(url=self.url.format(end_date, pn, ps))
        return resp.json()

    def run(self):
        current_date = get_recent_report_date()
        pn = 1
        ps = 2000

        while to_pd_timestamp(current_date) >= to_pd_timestamp(self.start):
            if not self.state:
                current_date = get_recent_report_date()
                result = self.get_data(end_date=current_date, pn=pn, ps=ps)
                print(result)
                self.state = {
                    'end_date': current_date,
                    'pages': result['result']['pages'],
                    'pn': pn,
                    'ps': ps
                }
                self.persist_state('stock_sz_000001', self.state)
            else:
                if self.state['pn'] >= self.state['pages']:
                    current_date = get_recent_report_date(the_date=self.state['end_date'], step=1)
                    pn = pn
                    ps = ps
                else:
                    pn = self.state['pn'] + 1
                    ps = self.state['ps']
                    current_date = self.state['end_date']

                result = self.get_data(end_date=current_date, pn=pn, ps=ps)
                print(result)
                self.state = {
                    'end_date': current_date,
                    'pages': result['result']['pages'],
                    'pn': pn,
                    'ps': ps
                }
                self.persist_state('stock_sz_000001', self.state)


if __name__ == '__main__':
    EastmoneyActorRecorder().run()
# the __all__ is generated
__all__ = ['EastmoneyActorRecorder']


# -*- coding: utf-8 -*-
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule decorator
from .decorator import *
from .decorator import __all__ as _decorator_all
__all__ += _decorator_all

# import all from submodule time_utils
from .time_utils import *
from .time_utils import __all__ as _time_utils_all
__all__ += _time_utils_all

# import all from submodule utils
from .utils import *
from .utils import __all__ as _utils_all
__all__ += _utils_all

# import all from submodule file_utils
from .file_utils import *
from .file_utils import __all__ as _file_utils_all
__all__ += _file_utils_all

# import all from submodule zip_utils
from .zip_utils import *
from .zip_utils import __all__ as _zip_utils_all
__all__ += _zip_utils_all

# import all from submodule git_utils
from .git_utils import *
from .git_utils import __all__ as _git_utils_all
__all__ += _git_utils_all

# import all from submodule pd_utils
from .pd_utils import *
from .pd_utils import __all__ as _pd_utils_all
__all__ += _pd_utils_all


# -*- coding: utf-8 -*-#
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule ma_stats_factor
from .ma_stats_factor import *
from .ma_stats_factor import __all__ as _ma_stats_factor_all
__all__ += _ma_stats_factor_all

# import all from submodule top_bottom_factor
from .top_bottom_factor import *
from .top_bottom_factor import __all__ as _top_bottom_factor_all
__all__ += _top_bottom_factor_all

# import all from submodule ma_factor
from .ma_factor import *
from .ma_factor import __all__ as _ma_factor_all
__all__ += _ma_factor_all

# import all from submodule domain
from .domain import *
from .domain import __all__ as _domain_all
__all__ += _domain_all


# -*- coding: utf-8 -*-

from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import Recorder
from zvtm.domain import Index
from zvtm.recorders.exchange.api import cn_index_api, cs_index_api


class ExchangeIndexRecorder(Recorder):
    provider = 'exchange'
    data_schema = Index

    def run(self):
        # 深圳
        self.record_cn_index('sz')
        # 国证
        self.record_cn_index('cni')

        # 上海
        self.record_cs_index('sh')
        # 中证
        self.record_cs_index('csi')

    # 中证，上海
    def record_cs_index(self, index_type):
        if index_type == 'csi':
            category_map_url = cs_index_api.csi_category_map_url
        elif index_type == 'sh':
            category_map_url = cs_index_api.sh_category_map_url
        else:
            self.logger.warning(f'not support index type: {index_type}')
            assert False

        for category, _ in category_map_url.items():
            df = cs_index_api.get_cs_index(index_type=index_type, category=category)
            df_to_db(data_schema=self.data_schema, df=df, provider=self.provider,
                     force_update=True)
            self.logger.info(f"finish record {index_type} index:{category.value}")

    # 国证，深圳
    def record_cn_index(self, index_type):
        if index_type == 'cni':
            category_map_url = cn_index_api.cni_category_map_url
        elif index_type == 'sz':
            category_map_url = cn_index_api.sz_category_map_url
        else:
            self.logger.error(f'not support index_type: {index_type}')
            assert False

        for category, _ in category_map_url.items():
            df = cn_index_api.get_cn_index(index_type=index_type, category=category)
            df_to_db(data_schema=self.data_schema, df=df, provider=self.provider,
                     force_update=True)
            self.logger.info(f"finish record {index_type} index:{category.value}")


if __name__ == '__main__':
    # init_log('china_stock_category.log')
    ExchangeIndexRecorder().run()

# the __all__ is generated
__all__ = ['ExchangeIndexRecorder']



# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule quotes
from .quotes import *
from .quotes import __all__ as _quotes_all
__all__ += _quotes_all

# import all from submodule money_flow
from .money_flow import *
from .money_flow import __all__ as _money_flow_all
__all__ += _money_flow_all

# import all from submodule meta
from .meta import *
from .meta import __all__ as _meta_all
__all__ += _meta_all


# -*- coding: utf-8 -*-
# this file is generated by gen_kdata_schema function, dont't change it
from sqlalchemy.orm import declarative_base

from zvtm.contract.register import register_schema
from zvtm.domain.quotes import StockhkKdataCommon

KdataBase = declarative_base()


class Stockhk1dHfqKdata(KdataBase, StockhkKdataCommon):
    __tablename__ = 'stockhk_1d_hfq_kdata'


register_schema(providers=['em'], db_name='stockhk_1d_hfq_kdata', schema_base=KdataBase, entity_type='stockhk')

# the __all__ is generated
__all__ = ['Stockhk1dHfqKdata']


# -*- coding: utf-8 -*-
import argparse

from zvtm.autocode import gen_exports
from zvtm.autocode.generator import gen_plugin_project


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--entity', help='entity name', default='future')
    parser.add_argument('--prefix', help='project prefix', default='zvt')
    parser.add_argument('--dir', help='project directory', default='.')
    parser.add_argument('--providers', help='providers', default=['joinquant'], nargs='+')

    args = parser.parse_args()

    dir_path = args.dir
    entity = args.entity
    providers = args.providers
    prefix = args.prefix
    gen_plugin_project(prefix=prefix, dir_path=dir_path, entity_type=entity, providers=providers)


def export():
    parser = argparse.ArgumentParser()
    parser.add_argument('--dir', help='export directory', default='.')
    args = parser.parse_args()
    dir_path = args.dir
    gen_exports(dir_path=dir_path)


if __name__ == '__main__':
    gen_plugin_project(dir_path='../../', entity_type='macro', providers=['zvt'])
    main()



# -*- coding: utf-8 -*-
import pandas as pd
from jqdatapy.api import run_query

from zvtm.api.portfolio import portfolio_relate_stock
from zvtm.api.utils import china_stock_code_to_id
from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import Recorder, TimeSeriesDataRecorder
from zvtm.domain.meta.fund_meta import Fund, FundStock
from zvtm.recorders.joinquant.common import to_entity_id, jq_to_report_period
from zvtm.utils.pd_utils import pd_is_not_null
from zvtm.utils.time_utils import to_time_str, next_date, now_pd_timestamp, is_same_date


class JqChinaFundRecorder(Recorder):
    provider = 'joinquant'
    data_schema = Fund

    def run(self):
        # 按不同类别抓取
        # 编码	基金运作方式
        # 401001	开放式基金
        # 401002	封闭式基金
        # 401003	QDII
        # 401004	FOF
        # 401005	ETF
        # 401006	LOF
        for operate_mode_id in (401001, 401002, 401005):
            year_count = 2
            while True:
                latest = Fund.query_data(filters=[Fund.operate_mode_id == operate_mode_id], order=Fund.timestamp.desc(),
                                         limit=1, return_type='domain')
                start_timestamp = '2000-01-01'
                if latest:
                    start_timestamp = latest[0].timestamp

                end_timestamp = min(next_date(start_timestamp, 365 * year_count), now_pd_timestamp())

                df = run_query(table='finance.FUND_MAIN_INFO',
                               conditions=f'operate_mode_id#=#{operate_mode_id}&start_date#>=#{to_time_str(start_timestamp)}&start_date#<=#{to_time_str(end_timestamp)}',
                               parse_dates=['start_date', 'end_date'],
                               dtype={'main_code': str})
                if not pd_is_not_null(df) or (df['start_date'].max().year < end_timestamp.year):
                    year_count = year_count + 1

                if pd_is_not_null(df):
                    df.rename(columns={'start_date': 'timestamp'}, inplace=True)
                    df['timestamp'] = pd.to_datetime(df['timestamp'])
                    df['list_date'] = df['timestamp']
                    df['end_date'] = pd.to_datetime(df['end_date'])

                    df['code'] = df['main_code']
                    df['entity_id'] = df['code'].apply(lambda x: to_entity_id(entity_type='fund', jq_code=x))
                    df['id'] = df['entity_id']
                    df['entity_type'] = 'fund'
                    df['exchange'] = 'sz'
                    df_to_db(df, data_schema=Fund, provider=self.provider, force_update=self.force_update)
                    self.logger.info(
                        f'persist fund {operate_mode_id} list success {start_timestamp} to {end_timestamp}')

                if is_same_date(end_timestamp, now_pd_timestamp()):
                    break


class JqChinaFundStockRecorder(TimeSeriesDataRecorder):
    entity_provider = 'joinquant'
    entity_schema = Fund

    provider = 'joinquant'
    data_schema = FundStock

    def init_entities(self):
        # 只抓股票型，混合型并且没退市的持仓,
        self.entities = Fund.query_data(
            entity_ids=self.entity_ids,
            codes=self.codes,
            return_type='domain',
            provider=self.entity_provider,
            filters=[Fund.underlying_asset_type.in_(('股票型', '混合型')), Fund.end_date.is_(None)])

    def record(self, entity, start, end, size, timestamps):
        # 忽略退市的
        if entity.end_date:
            return None
        redundant_times = 1
        while redundant_times > 0:
            df = run_query(table='finance.FUND_PORTFOLIO_STOCK',
                           conditions=f'pub_date#>=#{to_time_str(start)}&code#=#{entity.code}',
                           parse_dates=None)
            df = df.dropna()
            if pd_is_not_null(df):
                # data format
                #          id    code period_start  period_end    pub_date  report_type_id report_type  rank  symbol  name      shares    market_cap  proportion
                # 0   8640569  159919   2018-07-01  2018-09-30  2018-10-26          403003        第三季度     1  601318  中国平安  19869239.0  1.361043e+09        7.09
                # 1   8640570  159919   2018-07-01  2018-09-30  2018-10-26          403003        第三季度     2  600519  贵州茅台    921670.0  6.728191e+08        3.50
                # 2   8640571  159919   2018-07-01  2018-09-30  2018-10-26          403003        第三季度     3  600036  招商银行  18918815.0  5.806184e+08        3.02
                # 3   8640572  159919   2018-07-01  2018-09-30  2018-10-26          403003        第三季度     4  601166  兴业银行  22862332.0  3.646542e+08        1.90
                df['timestamp'] = pd.to_datetime(df['pub_date'])

                df.rename(columns={'symbol': 'stock_code', 'name': 'stock_name'}, inplace=True)
                df['proportion'] = df['proportion'] * 0.01

                df = portfolio_relate_stock(df, entity)

                df['stock_id'] = df['stock_code'].apply(lambda x: china_stock_code_to_id(x))
                df['id'] = df[['entity_id', 'stock_id', 'pub_date', 'id']].apply(lambda x: '_'.join(x.astype(str)),
                                                                                 axis=1)
                df['report_date'] = pd.to_datetime(df['period_end'])
                df['report_period'] = df['report_type'].apply(lambda x: jq_to_report_period(x))

                saved = df_to_db(df=df, data_schema=self.data_schema, provider=self.provider,
                                 force_update=self.force_update)

                # 取不到非重复的数据
                if saved == 0:
                    return None

                # self.logger.info(df.tail())
                self.logger.info(
                    f"persist fund {entity.code}({entity.name}) portfolio success {df.iloc[-1]['pub_date']}")
                latest = df['timestamp'].max()

                # 取到了最近两年的数据，再请求一次,确保取完最新的数据
                if latest.year >= now_pd_timestamp().year - 1:
                    redundant_times = redundant_times - 1
                start = latest
            else:
                return None

        return None


if __name__ == '__main__':
    # JqChinaFundRecorder().run()
    JqChinaFundStockRecorder(codes=['000053']).run()
# the __all__ is generated
__all__ = ['JqChinaFundRecorder', 'JqChinaFundStockRecorder']


# -*- coding: utf-8 -*-
import json

import demjson
import pandas as pd
import requests

from zvtm.api.utils import china_stock_code_to_id
from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import Recorder, TimeSeriesDataRecorder
from zvtm.domain import BlockStock, BlockCategory, Block
from zvtm.utils.time_utils import now_pd_timestamp


class SinaBlockRecorder(Recorder):
    provider = 'sina'
    data_schema = Block

    # 用于抓取行业/概念/地域列表
    category_map_url = {
        BlockCategory.industry: 'http://vip.stock.finance.sina.com.cn/q/view/newSinaHy.php',
        BlockCategory.concept: 'http://money.finance.sina.com.cn/q/view/newFLJK.php?param=class'
        # StockCategory.area: 'http://money.finance.sina.com.cn/q/view/newFLJK.php?param=area',
    }

    def run(self):
        # get stock blocks from sina
        for category, url in self.category_map_url.items():
            resp = requests.get(url)
            resp.encoding = 'GBK'

            tmp_str = resp.text
            json_str = tmp_str[tmp_str.index('{'):tmp_str.index('}') + 1]
            tmp_json = json.loads(json_str)

            the_list = []

            for code in tmp_json:
                name = tmp_json[code].split(',')[1]
                entity_id = f'block_cn_{code}'
                the_list.append({
                    'id': entity_id,
                    'entity_id': entity_id,
                    'entity_type': 'block',
                    'exchange': 'cn',
                    'code': code,
                    'name': name,
                    'category': category.value
                })
            if the_list:
                df = pd.DataFrame.from_records(the_list)
                df_to_db(data_schema=self.data_schema, df=df, provider=self.provider,
                         force_update=True)

            self.logger.info(f"finish record sina blocks:{category.value}")


class SinaChinaBlockStockRecorder(TimeSeriesDataRecorder):
    entity_provider = 'sina'
    entity_schema = Block

    provider = 'sina'
    data_schema = BlockStock

    # 用于抓取行业包含的股票
    category_stocks_url = 'http://vip.stock.finance.sina.com.cn/quotes_service/api/json_v2.php/Market_Center.getHQNodeData?page={}&num=5000&sort=symbol&asc=1&node={}&symbol=&_s_r_a=page'

    def record(self, entity, start, end, size, timestamps):
        for page in range(1, 5):
            resp = requests.get(self.category_stocks_url.format(page, entity.code))
            try:
                if resp.text == 'null' or resp.text is None:
                    break
                category_jsons = demjson.decode(resp.text)
                the_list = []
                for category in category_jsons:
                    stock_code = category['code']
                    stock_id = china_stock_code_to_id(stock_code)
                    block_id = entity.id
                    the_list.append({
                        'id': '{}_{}'.format(block_id, stock_id),
                        'entity_id': block_id,
                        'entity_type': 'block',
                        'exchange': entity.exchange,
                        'code': entity.code,
                        'name': entity.name,
                        'timestamp': now_pd_timestamp(),
                        'stock_id': stock_id,
                        'stock_code': stock_code,
                        'stock_name': category['name'],

                    })
                if the_list:
                    df = pd.DataFrame.from_records(the_list)
                    print(df)
                    df_to_db(data_schema=self.data_schema, df=df, provider=self.provider,
                             force_update=True)

                self.logger.info('finish recording BlockStock:{},{}'.format(entity.category, entity.name))

            except Exception as e:
                self.logger.error("error:,resp.text:", e, resp.text)
            self.sleep()


if __name__ == '__main__':
    # init_log('sina_china_stock_category.log')
    Block.record_data(provider='sina')
    recorder = SinaChinaBlockStockRecorder(codes=['new_cbzz'])
    recorder.run()
# the __all__ is generated
__all__ = ['SinaBlockRecorder', 'SinaChinaBlockStockRecorder']


# -*- coding: utf-8 -*-
import pandas as pd

from zvtm.api.portfolio import get_etf_stocks
from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import TimeSeriesDataRecorder
from zvtm.domain import StockValuation, Etf, EtfValuation
from zvtm.utils.pd_utils import pd_is_not_null
from zvtm.utils.time_utils import now_pd_timestamp


class JqChinaEtfValuationRecorder(TimeSeriesDataRecorder):
    entity_provider = 'joinquant'
    entity_schema = Etf

    # 数据来自jq
    provider = 'joinquant'

    data_schema = EtfValuation

    def record(self, entity, start, end, size, timestamps):
        if not end:
            end = now_pd_timestamp()

        date_range = pd.date_range(start=start, end=end, freq='1D').tolist()
        for date in date_range:
            # etf包含的个股和比例
            etf_stock_df = get_etf_stocks(code=entity.code, timestamp=date, provider=self.provider)

            if pd_is_not_null(etf_stock_df):
                all_pct = etf_stock_df['proportion'].sum()

                if all_pct >= 1.2 or all_pct <= 0.8:
                    self.logger.error(f'ignore etf:{entity.id}  date:{date} proportion sum:{all_pct}')
                    break

                etf_stock_df.set_index('stock_id', inplace=True)

                # 个股的估值数据
                stock_valuation_df = StockValuation.query_data(entity_ids=etf_stock_df.index.to_list(),
                                                               filters=[StockValuation.timestamp == date],
                                                               index='entity_id')

                if pd_is_not_null(stock_valuation_df):
                    stock_count = len(etf_stock_df)
                    valuation_count = len(stock_valuation_df)

                    self.logger.info(
                        f'etf:{entity.id} date:{date} stock count: {stock_count},'
                        f'valuation count:{valuation_count}')

                    pct = abs(stock_count - valuation_count) / stock_count

                    if pct >= 0.2:
                        self.logger.error(f'ignore etf:{entity.id}  date:{date} pct:{pct}')
                        break

                    se = pd.Series({'id': "{}_{}".format(entity.id, date),
                                    'entity_id': entity.id,
                                    'timestamp': date,
                                    'code': entity.code,
                                    'name': entity.name})
                    for col in ['pe', 'pe_ttm', 'pb', 'ps', 'pcf']:
                        # PE=P/E
                        # 这里的算法为：将其价格都设为PE,那么Earning为1(亏钱为-1)，结果为 总价格(PE)/总Earning

                        value = 0
                        price = 0

                        # 权重估值
                        positive_df = stock_valuation_df[[col]][stock_valuation_df[col] > 0]
                        positive_df['count'] = 1
                        positive_df = positive_df.multiply(etf_stock_df["proportion"], axis="index")
                        if pd_is_not_null(positive_df):
                            value = positive_df['count'].sum()
                            price = positive_df[col].sum()

                        negative_df = stock_valuation_df[[col]][stock_valuation_df[col] < 0]
                        if pd_is_not_null(negative_df):
                            negative_df['count'] = 1
                            negative_df = negative_df.multiply(etf_stock_df["proportion"], axis="index")
                            value = value - negative_df['count'].sum()
                            price = price + negative_df[col].sum()

                        se[f'{col}1'] = price / value

                        # 简单算术平均估值
                        positive_df = stock_valuation_df[col][stock_valuation_df[col] > 0]
                        positive_count = len(positive_df)

                        negative_df = stock_valuation_df[col][stock_valuation_df[col] < 0]
                        negative_count = len(negative_df)

                        value = positive_count - negative_count
                        price = positive_df.sum() + abs(negative_df.sum())

                        se[col] = price / value
                    df = se.to_frame().T

                    self.logger.info(df)

                    df_to_db(df=df, data_schema=self.data_schema, provider=self.provider,
                             force_update=self.force_update)

        return None


if __name__ == '__main__':
    # 上证50
    JqChinaEtfValuationRecorder(codes=['512290']).run()
# the __all__ is generated
__all__ = ['JqChinaEtfValuationRecorder']


# -*- coding: utf-8 -*-

import demjson
import pandas as pd
import requests

from zvtm import init_log
from zvtm.api.kdata import generate_kdata_id, get_kdata
from zvtm.contract import IntervalLevel
from zvtm.contract.recorder import FixedCycleDataRecorder
from zvtm.domain import Etf, Etf1dKdata
from zvtm.recorders.consts import EASTMONEY_ETF_NET_VALUE_HEADER
from zvtm.utils.time_utils import to_time_str


class ChinaETFDayKdataRecorder(FixedCycleDataRecorder):
    entity_provider = 'exchange'
    entity_schema = Etf

    provider = 'sina'
    data_schema = Etf1dKdata
    url = 'http://money.finance.sina.com.cn/quotes_service/api/json_v2.php/CN_MarketData.getKLineData?' \
          'symbol={}{}&scale=240&&datalen={}&ma=no'

    def get_data_map(self):
        return {}

    def generate_domain_id(self, entity, original_data):
        return generate_kdata_id(entity_id=entity.id, timestamp=original_data['timestamp'], level=self.level)

    def on_finish_entity(self, entity):
        kdatas = get_kdata(entity_id=entity.id, level=IntervalLevel.LEVEL_1DAY.value,
                           order=Etf1dKdata.timestamp.asc(),
                           return_type='domain', session=self.session,
                           filters=[Etf1dKdata.cumulative_net_value.is_(None)])

        if kdatas and len(kdatas) > 0:
            start = kdatas[0].timestamp
            end = kdatas[-1].timestamp

            # 从东方财富获取基金累计净值
            df = self.fetch_cumulative_net_value(entity, start, end)

            if df is not None and not df.empty:
                for kdata in kdatas:
                    if kdata.timestamp in df.index:
                        kdata.cumulative_net_value = df.loc[kdata.timestamp, 'LJJZ']
                        kdata.change_pct = df.loc[kdata.timestamp, 'JZZZL']
                self.session.commit()
                self.logger.info(f'{entity.code} - {entity.name}累计净值更新完成...')

    def fetch_cumulative_net_value(self, security_item, start, end) -> pd.DataFrame:
        query_url = 'http://api.fund.eastmoney.com/f10/lsjz?' \
                    'fundCode={}&pageIndex={}&pageSize=200&startDate={}&endDate={}'

        page = 1
        df = pd.DataFrame()
        while True:
            url = query_url.format(security_item.code, page, to_time_str(start), to_time_str(end))

            response = requests.get(url, headers=EASTMONEY_ETF_NET_VALUE_HEADER)
            response_json = demjson.decode(response.text)
            response_df = pd.DataFrame(response_json['Data']['LSJZList'])

            # 最后一页
            if response_df.empty:
                break

            response_df['FSRQ'] = pd.to_datetime(response_df['FSRQ'])
            response_df['JZZZL'] = pd.to_numeric(response_df['JZZZL'], errors='coerce')
            response_df['LJJZ'] = pd.to_numeric(response_df['LJJZ'], errors='coerce')
            response_df = response_df.fillna(0)
            response_df.set_index('FSRQ', inplace=True, drop=True)

            df = pd.concat([df, response_df])
            page += 1

            self.sleep()

        return df

    def record(self, entity, start, end, size, timestamps):
        # 此 url 不支持分页，如果超过我们想取的条数，则只能取最大条数
        if start is None or size > self.default_size:
            size = 8000

        param = {
            'security_item': entity,
            'level': self.level.value,
            'size': size
        }

        security_item = param['security_item']
        size = param['size']

        url = ChinaETFDayKdataRecorder.url.format(security_item.exchange, security_item.code, size)

        response = requests.get(url)
        response_json = demjson.decode(response.text)

        if response_json is None or len(response_json) == 0:
            return []

        df = pd.DataFrame(response_json)
        df.rename(columns={'day': 'timestamp'}, inplace=True)
        df['timestamp'] = pd.to_datetime(df['timestamp'])
        df['name'] = security_item.name
        df['provider'] = 'sina'
        df['level'] = param['level']

        return df.to_dict(orient='records')


__all__ = ['ChinaETFDayKdataRecorder']

if __name__ == '__main__':
    init_log('sina_china_etf_day_kdata.log')
    ChinaETFDayKdataRecorder(level=IntervalLevel.LEVEL_1DAY).run()
# the __all__ is generated
__all__ = ['ChinaETFDayKdataRecorder']


import demjson
import pandas as pd
import requests

from zvtm.contract.recorder import TimestampsDataRecorder
from zvtm.domain import Index
from zvtm.domain.misc import StockSummary
from zvtm.recorders.consts import DEFAULT_SH_SUMMARY_HEADER
from zvtm.utils.time_utils import to_time_str
from zvtm.utils.utils import to_float


class ExchangeStockSummaryRecorder(TimestampsDataRecorder):
    entity_provider = 'exchange'
    entity_schema = Index

    provider = 'exchange'
    data_schema = StockSummary

    original_page_url = 'http://www.sse.com.cn/market/stockdata/overview/day/'

    url = 'http://query.sse.com.cn/marketdata/tradedata/queryTradingByProdTypeData.do?jsonCallBack=jsonpCallback30731&searchDate={}&prodType=gp&_=1515717065511'

    def __init__(self, force_update=False, sleeping_time=5, exchanges=None, entity_ids=None, day_data=False,
                 entity_filters=None, ignore_failed=True, real_time=False, fix_duplicate_way='add',
                 start_timestamp=None, end_timestamp=None) -> None:
        super().__init__(force_update, sleeping_time, exchanges, entity_ids, ['000001'], day_data, entity_filters,
                         ignore_failed, real_time, fix_duplicate_way, start_timestamp, end_timestamp)

    def init_timestamps(self, entity):
        return pd.date_range(start=entity.timestamp,
                             end=pd.Timestamp.now(),
                             freq='B').tolist()

    def record(self, entity, start, end, size, timestamps):
        json_results = []
        for timestamp in timestamps:
            timestamp_str = to_time_str(timestamp)
            url = self.url.format(timestamp_str)
            response = requests.get(url=url, headers=DEFAULT_SH_SUMMARY_HEADER)

            results = demjson.decode(response.text[response.text.index("(") + 1:response.text.index(")")])['result']
            result = [result for result in results if result['productType'] == '1']
            if result and len(result) == 1:
                result_json = result[0]
                # 有些较老的数据不存在,默认设为0.0
                json_results.append({
                    'provider': 'exchange',
                    'timestamp': timestamp,
                    'name': '上证指数',
                    'pe': to_float(result_json['profitRate'], 0.0),
                    'total_value': to_float(result_json['marketValue1'] + '亿', 0.0),
                    'total_tradable_vaule': to_float(result_json['negotiableValue1'] + '亿', 0.0),
                    'volume': to_float(result_json['trdVol1'] + '万', 0.0),
                    'turnover': to_float(result_json['trdAmt1'] + '亿', 0.0),
                    'turnover_rate': to_float(result_json['exchangeRate'], 0.0),
                })

                if len(json_results) > 30:
                    return json_results

        return json_results

    def get_data_map(self):
        return None


if __name__ == '__main__':
    ExchangeStockSummaryRecorder().run()
# the __all__ is generated
__all__ = ['ExchangeStockSummaryRecorder']


# -*- coding: utf-8 -*-
import logging
from typing import Type

import pandas as pd

from zvtm.contract import Mixin
from zvtm.contract import TradableEntity
from zvtm.contract.api import get_db_session
from zvtm.contract.base import StatefulService
from zvtm.contract.zvt_info import TaggerState
from zvtm.domain import Stock
from zvtm.tag.dataset.stock_tags import StockTags
from zvtm.utils import to_time_str, to_pd_timestamp
from zvtm.utils.time_utils import TIME_FORMAT_DAY, now_pd_timestamp

logger = logging.getLogger(__name__)


class Tagger(StatefulService):
    state_schema = TaggerState

    entity_schema: Type[TradableEntity] = None

    data_schema: Type[Mixin] = None

    start_timestamp = '2005-01-01'

    def __init__(self, force=False) -> None:
        super().__init__()
        assert self.entity_schema is not None
        assert self.data_schema is not None
        self.force = force
        self.session = get_db_session(provider='zvt', data_schema=self.data_schema)
        if self.state and not self.force:
            logger.info(f'get start_timestamp from state')
            self.start_timestamp = self.state['current_timestamp']
        logger.info(f'tag start_timestamp: {self.start_timestamp}')

    def tag(self, timestamp):
        raise NotImplementedError

    def get_tag_timestamps(self):
        return pd.date_range(start=self.start_timestamp, end=now_pd_timestamp(), freq='M')

    def get_tag_domain(self, entity_id, timestamp, **fill_kv):
        the_date = to_time_str(timestamp, fmt=TIME_FORMAT_DAY)
        the_id = f'{entity_id}_{the_date}'
        the_domain = self.data_schema.get_one(id=the_id)

        if the_domain:
            for k, v in fill_kv.items():
                exec(f'the_domain.{k}=v')
        else:
            return self.data_schema(id=the_id,
                                    entity_id=entity_id,
                                    timestamp=to_pd_timestamp(the_date),
                                    **fill_kv)
        return the_domain

    def get_tag_domains(self, entity_ids, timestamp, **fill_kv):
        the_date = to_time_str(timestamp, fmt=TIME_FORMAT_DAY)
        ids = [f'{entity_id}_{the_date}' for entity_id in entity_ids]

        the_domains = self.data_schema.query_data(ids=ids, return_type='domain')

        if the_domains:
            for the_domain in the_domains:
                for k, v in fill_kv.items():
                    exec(f'the_domain.{k}=v')

        current_ids = [item.id for item in the_domains]
        need_new_ids = set(ids) - set(current_ids)
        new_domains = [
            self.data_schema(id=f'{entity_id}_{the_date}',
                             entity_id=entity_id,
                             timestamp=to_pd_timestamp(the_date),
                             **fill_kv) for
            entity_id in need_new_ids]
        return the_domains + new_domains

    def run(self):
        timestamps = self.get_tag_timestamps()
        for timestamp in timestamps:
            logger.info(f'tag to {timestamp}')
            self.tag(timestamp=timestamp)

            self.state = {'current_timestamp': to_time_str(timestamp)}
            self.persist_state()


class StockTagger(Tagger):
    data_schema = StockTags
    entity_schema = Stock

    def tag(self, timestamp):
        raise NotImplementedError



# -*- coding: utf-8 -*-
import logging
import time
from typing import List, Union, Type, Tuple

import pandas as pd

from zvtm.api.trader_info_api import AccountStatsReader
from zvtm.contract import IntervalLevel, TradableEntity, AdjustType
from zvtm.contract.drawer import Drawer
from zvtm.contract.normal_data import NormalData
from zvtm.domain import Stock, AccountStats, Position
from zvtm.factors.target_selector import TargetSelector
from zvtm.trader import TradingSignal, TradingSignalType, TradingListener
from zvtm.trader.account import SimAccountService
from zvtm.utils.time_utils import to_pd_timestamp, now_pd_timestamp, to_time_str, is_same_date


class Trader(object):
    entity_schema: Type[TradableEntity] = None

    def __init__(self,
                 entity_ids: List[str] = None,
                 exchanges: List[str] = None,
                 codes: List[str] = None,
                 start_timestamp: Union[str, pd.Timestamp] = None,
                 end_timestamp: Union[str, pd.Timestamp] = None,
                 provider: str = None,
                 level: Union[str, IntervalLevel] = IntervalLevel.LEVEL_1DAY,
                 trader_name: str = None,
                 real_time: bool = False,
                 kdata_use_begin_time: bool = False,
                 draw_result: bool = True,
                 rich_mode: bool = False,
                 adjust_type: AdjustType = None,
                 profit_threshold=(3, -0.3),
                 keep_history=False) -> None:
        assert self.entity_schema is not None
        assert start_timestamp is not None
        assert end_timestamp is not None

        self.logger = logging.getLogger(__name__)

        if trader_name:
            self.trader_name = trader_name
        else:
            self.trader_name = type(self).__name__.lower()

        self.entity_ids = entity_ids
        self.exchanges = exchanges
        self.codes = codes
        self.provider = provider
        # make sure the min level selector correspond to the provider and level
        self.level = IntervalLevel(level)
        self.real_time = real_time
        self.start_timestamp = to_pd_timestamp(start_timestamp)
        self.end_timestamp = to_pd_timestamp(end_timestamp)

        self.trading_dates = self.entity_schema.get_trading_dates(start_date=self.start_timestamp,
                                                                  end_date=self.end_timestamp)

        if real_time:
            self.logger.info(
                'real_time mode, end_timestamp should be future,you could set it big enough for running forever')
            assert self.end_timestamp >= now_pd_timestamp()

        self.kdata_use_begin_time = kdata_use_begin_time
        self.draw_result = draw_result
        self.rich_mode = rich_mode

        if type(adjust_type) is str:
            adjust_type = AdjustType(adjust_type)
        self.adjust_type = adjust_type
        self.profit_threshold = profit_threshold
        self.keep_history = keep_history

        self.level_map_long_targets = {}
        self.level_map_short_targets = {}
        self.trading_signals: List[TradingSignal] = []
        self.trading_signal_listeners: List[TradingListener] = []
        self.selectors: List[TargetSelector] = []

        self.account_service = SimAccountService(entity_schema=self.entity_schema,
                                                 trader_name=self.trader_name,
                                                 timestamp=self.start_timestamp,
                                                 provider=self.provider,
                                                 level=self.level,
                                                 rich_mode=self.rich_mode,
                                                 adjust_type=self.adjust_type,
                                                 keep_history=self.keep_history)

        self.register_trading_signal_listener(self.account_service)

        self.init_selectors(entity_ids=self.entity_ids, entity_schema=self.entity_schema, exchanges=self.exchanges,
                            codes=self.codes, start_timestamp=self.start_timestamp, end_timestamp=self.end_timestamp,
                            adjust_type=self.adjust_type)

        if self.selectors:
            self.trading_level_asc = list(set([IntervalLevel(selector.level) for selector in self.selectors]))
            self.trading_level_asc.sort()

            self.logger.info(f'trader level:{self.level},selectors level:{self.trading_level_asc}')

            if self.level != self.trading_level_asc[0]:
                raise Exception("trader level should be the min of the selectors")

            self.trading_level_desc = list(self.trading_level_asc)
            self.trading_level_desc.reverse()

            # run selectors for history data at first
            for selector in self.selectors:
                selector.run()

        self.on_start()

    def on_start(self):
        self.logger.info(f'trader:{self.trader_name} on_start')

    def init_selectors(self, entity_ids, entity_schema, exchanges, codes, start_timestamp, end_timestamp,
                       adjust_type=None):
        """
        overwrite it to init selectors if you want to use selector/factor computing model
        :param adjust_type:

        """
        pass

    def update_targets_by_level(self, level: IntervalLevel, long_targets: List[str],
                                short_targets: List[str], ) -> None:
        """
        the trading signals is generated in min level,before that,we should cache targets of all levels

        :param level:
        :param long_targets:
        :param short_targets:
        """
        self.logger.debug(
            f'level:{level},old long targets:{self.level_map_long_targets.get(level)},new long targets:{long_targets}')
        self.level_map_long_targets[level] = long_targets

        self.logger.debug(
            f'level:{level},old short targets:{self.level_map_short_targets.get(level)},new short targets:{short_targets}')
        self.level_map_short_targets[level] = short_targets

    def get_long_targets_by_level(self, level: IntervalLevel) -> List[str]:
        return self.level_map_long_targets.get(level)

    def get_short_targets_by_level(self, level: IntervalLevel) -> List[str]:
        return self.level_map_short_targets.get(level)

    def on_targets_selected_from_levels(self, timestamp) -> Tuple[List[str], List[str]]:
        """
        this method's called in every min level cycle to select targets in all levels generated by the previous cycle
        the default implementation is selecting the targets in all levels
        overwrite it for your custom logic

        :param timestamp: current event time
        :return: long targets, short targets
        """

        long_selected = None

        short_selected = None

        for level in self.trading_level_desc:
            long_targets = self.level_map_long_targets.get(level)
            # long must in all
            if long_targets:
                long_targets = set(long_targets)
                if long_selected is None:
                    long_selected = long_targets
                else:
                    long_selected = long_selected & long_targets
            else:
                long_selected = set()

            short_targets = self.level_map_short_targets.get(level)
            # short any
            if short_targets:
                short_targets = set(short_targets)
                if short_selected is None:
                    short_selected = short_targets
                else:
                    short_selected = short_selected | short_targets

        return long_selected, short_selected

    def get_current_account(self) -> AccountStats:
        return self.account_service.get_current_account()

    def get_current_positions(self) -> List[Position]:
        return self.get_current_account().positions

    def long_position_control(self):
        positions = self.get_current_positions()

        position_pct = 1.0
        if not positions:
            # 没有仓位，买2成
            position_pct = 0.2
        elif len(positions) <= 10:
            # 小于10个持仓，买5成
            position_pct = 0.5

        # 买完
        return position_pct

    def short_position_control(self):
        # 卖完
        return 1.0

    def on_profit_control(self):
        if self.profit_threshold and self.get_current_positions():
            positive = self.profit_threshold[0]
            negative = self.profit_threshold[1]
            close_long_entity_ids = []
            for position in self.get_current_positions():
                if position.available_long > 1:
                    # 止盈
                    if position.profit_rate >= positive:
                        close_long_entity_ids.append(position.entity_id)
                        self.logger.info(f'close profit {position.profit_rate} for {position.entity_id}')
                    # 止损
                    if position.profit_rate <= negative:
                        close_long_entity_ids.append(position.entity_id)
                        self.logger.info(f'cut lost {position.profit_rate} for {position.entity_id}')

            return close_long_entity_ids, None
        return None, None

    def buy(self, due_timestamp, happen_timestamp, entity_ids, ignore_in_position=True):
        if ignore_in_position:
            account = self.get_current_account()
            current_holdings = []
            if account.positions:
                current_holdings = [position.entity_id for position in account.positions if position != None and
                                    position.available_long > 0]

            entity_ids = set(entity_ids) - set(current_holdings)

        if entity_ids:
            position_pct = self.long_position_control()
            position_pct = (1.0 / len(entity_ids)) * position_pct

            for entity_id in entity_ids:
                trading_signal = TradingSignal(entity_id=entity_id,
                                               due_timestamp=due_timestamp,
                                               happen_timestamp=happen_timestamp,
                                               trading_signal_type=TradingSignalType.open_long,
                                               trading_level=self.level,
                                               position_pct=position_pct)
                self.trading_signals.append(trading_signal)

    def sell(self, due_timestamp, happen_timestamp, entity_ids):
        # current position
        account = self.get_current_account()
        current_holdings = []
        if account.positions:
            current_holdings = [position.entity_id for position in account.positions if position != None and
                                position.available_long > 0]

        shorted = set(current_holdings) & set(entity_ids)

        if shorted:
            position_pct = self.short_position_control()

            for entity_id in shorted:
                trading_signal = TradingSignal(entity_id=entity_id,
                                               due_timestamp=due_timestamp,
                                               happen_timestamp=happen_timestamp,
                                               trading_signal_type=TradingSignalType.close_long,
                                               trading_level=self.level,
                                               position_pct=position_pct)
                self.trading_signals.append(trading_signal)

    def trade_the_targets(self, due_timestamp, happen_timestamp, long_selected, short_selected):
        if short_selected:
            self.sell(due_timestamp=due_timestamp, happen_timestamp=happen_timestamp, entity_ids=short_selected)
        if long_selected:
            self.buy(due_timestamp=due_timestamp, happen_timestamp=happen_timestamp, entity_ids=long_selected)

    def on_finish(self, timestmap):
        self.on_trading_finish(timestmap)
        # show the result
        if self.draw_result:
            import plotly.io as pio
            pio.renderers.default = "browser"
            reader = AccountStatsReader(trader_names=[self.trader_name])
            df = reader.data_df
            drawer = Drawer(main_data=NormalData(df.copy()[['trader_name', 'timestamp', 'all_value']],
                                                 category_field='trader_name'))
            drawer.draw_line(show=True)

    def on_targets_filtered(self, timestamp, level, selector: TargetSelector, long_targets: List[str],
                            short_targets: List[str]) -> Tuple[List[str], List[str]]:
        """
        overwrite it to filter the targets from selector

        :param timestamp: the event time
        :param level: the level
        :param selector: the selector
        :param long_targets: the long targets from the selector
        :param short_targets: the short targets from the selector
        :return: filtered long targets, filtered short targets
        """
        self.logger.info(f'on_targets_filtered {level} long:{long_targets}')

        if len(long_targets) > 10:
            long_targets = long_targets[0:10]
        self.logger.info(f'on_targets_filtered {level} filtered long:{long_targets}')

        return long_targets, short_targets

    def in_trading_date(self, timestamp):
        return to_time_str(timestamp) in self.trading_dates

    def on_time(self, timestamp: pd.Timestamp):
        """
        called in every min level cycle

        :param timestamp: event time
        """
        self.logger.debug(f'current timestamp:{timestamp}')

    def on_trading_signals(self, trading_signals: List[TradingSignal]):
        for l in self.trading_signal_listeners:
            l.on_trading_signals(trading_signals)

    def on_trading_signal(self, trading_signal: TradingSignal):
        for l in self.trading_signal_listeners:
            try:
                l.on_trading_signal(trading_signal)
            except Exception as e:
                self.logger.exception(e)
                l.on_trading_error(timestamp=trading_signal.happen_timestamp, error=e)

    def on_trading_open(self, timestamp):
        for l in self.trading_signal_listeners:
            l.on_trading_open(timestamp)

    def on_trading_close(self, timestamp):
        for l in self.trading_signal_listeners:
            l.on_trading_close(timestamp)

    def on_trading_finish(self, timestamp):
        for l in self.trading_signal_listeners:
            l.on_trading_finish(timestamp)

    def on_trading_error(self, timestamp, error):
        for l in self.trading_signal_listeners:
            l.on_trading_error(timestamp, error)

    def run(self):
        # iterate timestamp of the min level,e.g,9:30,9:35,9.40...for 5min level
        # timestamp represents the timestamp in kdata
        for timestamp in self.entity_schema.get_interval_timestamps(start_date=self.start_timestamp,
                                                                    end_date=self.end_timestamp, level=self.level):

            if not self.in_trading_date(timestamp=timestamp):
                continue

            waiting_seconds = 0

            if self.level == IntervalLevel.LEVEL_1DAY:
                if is_same_date(timestamp, now_pd_timestamp()):
                    while True:
                        self.logger.info(f'time is:{now_pd_timestamp()},just smoke for minutes')
                        time.sleep(60)
                        current = now_pd_timestamp()
                        if current.hour >= 19:
                            waiting_seconds = 20
                            break

            elif self.real_time:
                # all selector move on to handle the coming data
                if self.kdata_use_begin_time:
                    real_end_timestamp = timestamp + pd.Timedelta(seconds=self.level.to_second())
                else:
                    real_end_timestamp = timestamp

                seconds = (now_pd_timestamp() - real_end_timestamp).total_seconds()
                waiting_seconds = self.level.to_second() - seconds

            # meaning the future kdata not ready yet,we could move on to check
            if waiting_seconds > 0:
                # iterate the selector from min to max which in finished timestamp kdata
                for level in self.trading_level_asc:
                    if self.entity_schema.is_finished_kdata_timestamp(timestamp=timestamp, level=level):
                        for selector in self.selectors:
                            if selector.level == level:
                                selector.move_on(timestamp, self.kdata_use_begin_time, timeout=waiting_seconds + 20)

            # on_trading_open to setup the account
            if self.level >= IntervalLevel.LEVEL_1DAY or (
                    self.level != IntervalLevel.LEVEL_1DAY and self.entity_schema.is_open_timestamp(timestamp)):
                self.on_trading_open(timestamp)

            self.on_time(timestamp=timestamp)

            # 一般来说selector(factors)计算 多标的 历史数据比较快，多级别的计算也比较方便，常用于全市场标的粗过滤
            # 更细节的控制可以在on_targets_filtered里进一步处理
            # 也可以在on_time里面设计一些自己的逻辑配合过滤
            if self.selectors:
                # 多级别的遍历算法要点:
                # 1)计算各级别的 标的，通过 on_targets_filtered 过滤，缓存在level_map_long_targets，level_map_short_targets
                # 2)在最小的level通过 on_targets_selected_from_levels 根据多级别的缓存标的，生成最终的选中标的
                # 这里需要注意的是，小级别拿到上一个周期的大级别的标的，这是合理的
                for level in self.trading_level_asc:
                    # in every cycle, all level selector do its job in its time
                    if self.entity_schema.is_finished_kdata_timestamp(timestamp=timestamp, level=level):
                        all_long_targets = []
                        all_short_targets = []

                        # 从该level的selector中过滤targets
                        for selector in self.selectors:
                            if selector.level == level:
                                long_targets = selector.get_open_long_targets(timestamp=timestamp)
                                short_targets = selector.get_open_short_targets(timestamp=timestamp)

                                if long_targets or short_targets:
                                    long_targets, short_targets = self.on_targets_filtered(timestamp=timestamp,
                                                                                           level=level,
                                                                                           selector=selector,
                                                                                           long_targets=long_targets,
                                                                                           short_targets=short_targets)

                                if long_targets:
                                    all_long_targets += long_targets
                                if short_targets:
                                    all_short_targets += short_targets

                        # 将各级别的targets缓存在level_map_long_targets，level_map_short_targets
                        self.update_targets_by_level(level, all_long_targets, all_short_targets)

                        # the time always move on by min level step and we could check all targets of levels
                        # 1)the targets is generated for next interval
                        # 2)the acceptable price is next interval prices,you could buy it at current price
                        # if the time is before the timestamp(due_timestamp) when trading signal received
                        # 3)the suggest price the the close price for generating the signal(happen_timestamp)
                        due_timestamp = timestamp + pd.Timedelta(seconds=self.level.to_second())

                        # 在最小level生成最终的 交易信号
                        if level == self.level:
                            long_selected, short_selected = self.on_targets_selected_from_levels(timestamp)

                            # 处理 止赢 止损
                            passive_short, _ = self.on_profit_control()
                            if passive_short:
                                if not short_selected:
                                    short_selected = passive_short
                                else:
                                    short_selected = list(set(short_selected) | set(passive_short))

                            self.logger.debug('timestamp:{},long_selected:{}'.format(due_timestamp, long_selected))
                            self.logger.debug('timestamp:{},short_selected:{}'.format(due_timestamp, short_selected))

                            if long_selected or short_selected:
                                self.trade_the_targets(due_timestamp=due_timestamp, happen_timestamp=timestamp,
                                                       long_selected=long_selected, short_selected=short_selected)

            if self.trading_signals:
                self.on_trading_signals(self.trading_signals)
            # clear
            self.trading_signals = []

            # on_trading_close to calculate date account
            if self.level >= IntervalLevel.LEVEL_1DAY or (
                    self.level != IntervalLevel.LEVEL_1DAY and self.entity_schema.is_close_timestamp(timestamp)):
                self.on_trading_close(timestamp)

        self.on_finish(timestamp)

    def register_trading_signal_listener(self, listener):
        if listener not in self.trading_signal_listeners:
            self.trading_signal_listeners.append(listener)

    def deregister_trading_signal_listener(self, listener):
        if listener in self.trading_signal_listeners:
            self.trading_signal_listeners.remove(listener)


class StockTrader(Trader):
    entity_schema = Stock

    def __init__(self, entity_ids: List[str] = None, exchanges: List[str] = None, codes: List[str] = None,
                 start_timestamp: Union[str, pd.Timestamp] = None, end_timestamp: Union[str, pd.Timestamp] = None,
                 provider: str = None, level: Union[str, IntervalLevel] = IntervalLevel.LEVEL_1DAY,
                 trader_name: str = None, real_time: bool = False, kdata_use_begin_time: bool = False,
                 draw_result: bool = True, rich_mode: bool = False, adjust_type: AdjustType = AdjustType.hfq,
                 profit_threshold=(3, -0.3), keep_history=False) -> None:
        super().__init__(entity_ids, exchanges, codes, start_timestamp, end_timestamp, provider, level, trader_name,
                         real_time, kdata_use_begin_time, draw_result, rich_mode, adjust_type, profit_threshold,
                         keep_history)


# the __all__ is generated
__all__ = ['Trader', 'StockTrader']



# -*- coding: utf-8 -*-
import logging
from enum import Enum

from zvtm.domain import IndexStock
from zvtm.tag.tag import StockTagger
from zvtm.utils import pd_is_not_null
from zvtm.utils.time_utils import month_start_date, month_end_date

logger = logging.getLogger(__name__)


class MarketValueTag(Enum):
    # 市值前500
    huge = 'huge'
    # 500-1000
    medium = 'medium'
    # 1000->
    small = 'small'


index_map_market_value = {
    # 399311 国证1000指数由A股市场中市值大、流动性好的1000只股票构成，反映A股市场大中盘股票的价格变动趋势。
    # 399400 大中盘指数由巨潮大盘指数样本股与巨潮中盘指数样本股组合而成,国证1000中的前500
    'index_sz_399400': MarketValueTag.huge,
    # 399316 巨潮小盘指数以国证1000指数为样本空间，选取总市值排名在后500名的股票
    'index_sz_399316': MarketValueTag.medium,
    # 399303 国证2000由全部A股扣除国证1000指数样本股后，市值大、流动性好的2000只A股组成，反映A股市场小微盘股票的价格变动趋势。
    'index_sz_399303': MarketValueTag.small
}


class MarketValueTagger(StockTagger):
    def tag(self, timestamp):
        for index_id in index_map_market_value:
            df = IndexStock.query_data(entity_id=index_id, start_timestamp=month_start_date(timestamp),
                                       end_timestamp=month_end_date(timestamp))
            if not pd_is_not_null(df):
                logger.error(f'no IndexStock data at {timestamp} for {index_id}')
                continue
            print(df)
            stock_tags = [self.get_tag_domain(entity_id=stock_id, timestamp=timestamp) for stock_id in
                          df['stock_id'].tolist()]

            for stock_tag in stock_tags:
                stock_tag.market_value_tag = index_map_market_value.get(index_id).value

            self.session.add_all(stock_tags)
            self.session.commit()


if __name__ == '__main__':
    MarketValueTagger().run()



# -*- coding: utf-8 -*-
# this file is generated by gen_kdata_schema function, dont't change it
from sqlalchemy.orm import declarative_base

from zvtm.contract.register import register_schema
from zvtm.domain.quotes import EtfKdataCommon

KdataBase = declarative_base()


class Etf1dKdata(KdataBase, EtfKdataCommon):
    __tablename__ = 'etf_1d_kdata'


register_schema(providers=['sina'], db_name='etf_1d_kdata', schema_base=KdataBase, entity_type='etf')

# the __all__ is generated
__all__ = ['Etf1dKdata']


# -*- coding: utf-8 -*-
from sqlalchemy import Column, Float, String, Boolean, Integer

from zvtm.contract import Mixin


class ZenFactorCommon(Mixin):
    level = Column(String(length=32))
    # 开盘价
    open = Column(Float)
    # 收盘价
    close = Column(Float)
    # 最高价
    high = Column(Float)
    # 最低价
    low = Column(Float)
    # 成交量
    volume = Column(Float)
    # 成交金额
    turnover = Column(Float)

    # 笔的底
    bi_di = Column(Boolean)
    # 笔的顶
    bi_ding = Column(Boolean)
    # 记录笔顶/底分型的值，bi_di取low,bi_ding取high,其他为None,绘图时取有值的连线即为 笔
    bi_value = Column(Float)
    # 笔的变化
    bi_change = Column(Float)
    # 笔的斜率
    bi_slope = Column(Float)
    # 持续的周期
    bi_interval = Column(Integer)

    # 记录临时分型，不变
    tmp_ding = Column(Boolean)
    tmp_di = Column(Boolean)
    # 分型的力度
    fenxing_power = Column(Float)

    # 目前分型确定的方向
    current_direction = Column(String(length=16))
    current_change = Column(Float)
    current_interval = Column(Integer)
    current_slope = Column(Float)
    # 最近的一个笔中枢
    # current_zhongshu = Column(String(length=512))
    current_zhongshu_y0 = Column(Float)
    current_zhongshu_y1 = Column(Float)
    current_zhongshu_change = Column(Float)

    current_merge_zhongshu_y0 = Column(Float)
    current_merge_zhongshu_y1 = Column(Float)
    current_merge_zhongshu_change = Column(Float)
    current_merge_zhongshu_level = Column(Integer)
    current_merge_zhongshu_interval = Column(Integer)

    # 目前走势的临时方向 其跟direction的的关系 确定了下一个分型
    tmp_direction = Column(String(length=16))
    opposite_change = Column(Float)
    opposite_slope = Column(Float)
    opposite_interval = Column(Integer)

    duan_state = Column(String(length=32))

    # 段的底
    duan_di = Column(Boolean)
    # 段的顶
    duan_ding = Column(Boolean)
    # 记录段顶/底的值，为duan_di时取low,为duan_ding时取high,其他为None,绘图时取有值的连线即为 段
    duan_value = Column(Float)
    # 段的变化
    duan_change = Column(Float)
    # 段的斜率
    duan_slope = Column(Float)
    # 持续的周期
    duan_interval = Column(Integer)

    # 记录在确定中枢的最后一个段的终点x1，值为Rect(x0,y0,x1,y1)
    zhongshu = Column(String(length=512))
    zhongshu_change = Column(Float)

    # 记录在确定中枢的最后一个笔的终点x1，值为Rect(x0,y0,x1,y1)
    bi_zhongshu = Column(String(length=512))
    bi_zhongshu_change = Column(Float)

    # 从前往后，合并相邻的有重叠的笔中枢
    merge_zhongshu = Column(String(length=512))
    merge_zhongshu_change = Column(Float)


# the __all__ is generated
__all__ = ['ZenFactorCommon']



# -*- coding: utf-8 -*-
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule cs_index_api
from .cs_index_api import *
from .cs_index_api import __all__ as _cs_index_api_all
__all__ += _cs_index_api_all

# import all from submodule cs_index_stock_api
from .cs_index_stock_api import *
from .cs_index_stock_api import __all__ as _cs_index_stock_api_all
__all__ += _cs_index_stock_api_all

# import all from submodule cn_index_api
from .cn_index_api import *
from .cn_index_api import __all__ as _cn_index_api_all
__all__ += _cn_index_api_all

# import all from submodule cn_index_stock_api
from .cn_index_stock_api import *
from .cn_index_stock_api import __all__ as _cn_index_stock_api_all
__all__ += _cn_index_stock_api_all


# -*- coding: utf-8 -*-
from zvtm.domain.fundamental.dividend_financing import DividendFinancing
from zvtm.recorders.eastmoney.common import EastmoneyPageabeDataRecorder
from zvtm.utils.utils import second_item_to_float


class DividendFinancingRecorder(EastmoneyPageabeDataRecorder):
    data_schema = DividendFinancing

    url = 'https://emh5.eastmoney.com/api/FenHongRongZi/GetLiNianFenHongRongZiList'
    page_url = url
    path_fields = ['LiNianFenHongRongZiList']

    def get_original_time_field(self):
        return 'ShiJian'

    def get_data_map(self):
        return {
            # 分红总额
            "dividend_money": ("FenHongZongE", second_item_to_float),
            # 新股
            "ipo_issues": ("XinGu", second_item_to_float),
            # 增发
            "spo_issues": ("ZengFa", second_item_to_float),
            # 配股
            "rights_issues": ("PeiFa", second_item_to_float)
        }

    def on_finish(self):
        try:
            code_security = {}
            for item in self.entities:
                code_security[item.code] = item

                need_fill_items = DividendFinancing.query_data(provider=self.provider, codes=list(code_security.keys()),
                                                               return_type='domain',
                                                               session=self.session,
                                                               filters=[
                                                                   DividendFinancing.ipo_raising_fund.is_(None),
                                                                   DividendFinancing.ipo_issues != 0])

                for need_fill_item in need_fill_items:
                    if need_fill_item:
                        need_fill_item.ipo_raising_fund = code_security[item.code].raising_fund
                        self.session.commit()
        except Exception as e:
            self.logger.exception(e)

        super().on_finish()


if __name__ == '__main__':
    # init_log('dividend_financing.log')

    recorder = DividendFinancingRecorder(codes=['000999'])
    recorder.run()
# the __all__ is generated
__all__ = ['DividendFinancingRecorder']


# script to auto generate some files
from zvtm.autocode.generator import gen_kdata_schema, gen_exports
from zvtm.contract import AdjustType
from zvtm.contract import IntervalLevel


def gen_kdata_schemas():
    # A股行情
    gen_kdata_schema(pkg='zvt', providers=['joinquant', 'em'], entity_type='stock',
                     levels=[level for level in IntervalLevel if level != IntervalLevel.LEVEL_TICK],
                     adjust_types=[None, AdjustType.hfq], entity_in_submodule=True)

    # 美股
    gen_kdata_schema(pkg='zvt', providers=['em'], entity_type='stockus',
                     levels=[IntervalLevel.LEVEL_1DAY],
                     adjust_types=[None, AdjustType.hfq], entity_in_submodule=True)

    # 港股
    gen_kdata_schema(pkg='zvt', providers=['em'], entity_type='stockhk',
                     levels=[IntervalLevel.LEVEL_1DAY],
                     adjust_types=[None, AdjustType.hfq], entity_in_submodule=True)

    # 板块行情
    gen_kdata_schema(pkg='zvt', providers=['em'], entity_type='block',
                     levels=[IntervalLevel.LEVEL_1DAY, IntervalLevel.LEVEL_1WEEK, IntervalLevel.LEVEL_1MON],
                     entity_in_submodule=True)

    # 指数行情
    gen_kdata_schema(pkg='zvt', providers=['em', 'sina'], entity_type='index',
                     levels=[IntervalLevel.LEVEL_1DAY, IntervalLevel.LEVEL_1WEEK], entity_in_submodule=True)

    # etf行情
    gen_kdata_schema(pkg='zvt', providers=['sina'], entity_type='etf',
                     levels=[IntervalLevel.LEVEL_1DAY], entity_in_submodule=True)


if __name__ == '__main__':
    # zip_dir(ZVT_TEST_DATA_PATH, zip_file_name=DATA_SAMPLE_ZIP_PATH)
    # gen_exports('contract',export_from_package=False)
    # gen_exports('recorders')
    # gen_exports('domain')
    # gen_exports('informer')
    gen_exports('utils')
    # gen_exports('trader')
    # gen_exports('autocode')
    # gen_exports('factors')
    # gen_kdata_schemas()



# -*- coding: utf-8 -*-
import logging
import numbers
from decimal import *
from enum import Enum
from urllib import parse

import pandas as pd

from zvtm.utils.time_utils import to_time_str

getcontext().prec = 16

logger = logging.getLogger(__name__)

none_values = ['不变', '--', '-', '新进']
zero_values = ['不变', '--', '-', '新进']


def first_item_to_float(the_list):
    return to_float(the_list[0])


def second_item_to_float(the_list):
    return to_float(the_list[1])


def add_func_to_value(the_map, the_func):
    for k, v in the_map.items():
        the_map[k] = (v, the_func)
    return the_map


def to_float(the_str, default=None):
    if not the_str:
        return default
    if the_str in none_values:
        return None

    if '%' in the_str:
        return pct_to_float(the_str)
    try:
        scale = 1.0
        if the_str[-2:] == '万亿':
            the_str = the_str[0:-2]
            scale = 1000000000000
        elif the_str[-1] == '亿':
            the_str = the_str[0:-1]
            scale = 100000000
        elif the_str[-1] == '万':
            the_str = the_str[0:-1]
            scale = 10000
        if not the_str:
            return default
        return float(Decimal(the_str.replace(',', '')) * Decimal(scale))
    except Exception as e:
        logger.error('the_str:{}'.format(the_str))
        logger.exception(e)
        return default


def pct_to_float(the_str, default=None):
    if the_str in none_values:
        return None

    try:
        return float(Decimal(the_str.replace('%', '')) / Decimal(100))
    except Exception as e:
        logger.exception(e)
        return default


def json_callback_param(the_str):
    json_str = the_str[the_str.index("(") + 1:the_str.rindex(")")].replace('null', 'None')
    return eval(json_str)


def fill_domain_from_dict(the_domain, the_dict: dict, the_map: dict = None, default_func=lambda x: x):
    """
    use field map and related func to fill properties from the dict to the domain


    :param the_domain:
    :type the_domain: DeclarativeMeta
    :param the_dict:
    :type the_dict: dict
    :param the_map:
    :type the_map: dict
    :param default_func:
    :type default_func: function
    """
    if not the_map:
        the_map = {}
        for k in the_dict:
            the_map[k] = (k, default_func)

    for k, v in the_map.items():
        if isinstance(v, tuple):
            field_in_dict = v[0]
            the_func = v[1]
        else:
            field_in_dict = v
            the_func = default_func

        the_value = the_dict.get(field_in_dict)
        if the_value is not None:
            to_value = the_value
            if to_value in none_values:
                setattr(the_domain, k, None)
            else:
                result_value = the_func(to_value)
                setattr(the_domain, k, result_value)
                exec('the_domain.{}=result_value'.format(k))


SUPPORT_ENCODINGS = ['GB2312', 'GBK', 'GB18030', 'UTF-8']


def read_csv(f, encoding, sep=None, na_values=None):
    encodings = [encoding] + SUPPORT_ENCODINGS
    for encoding in encodings:
        try:
            if sep:
                return pd.read_csv(f, sep=sep, encoding=encoding, na_values=na_values)
            else:
                return pd.read_csv(f, encoding=encoding, na_values=na_values)
        except UnicodeDecodeError as e:
            logger.warning('read_csv failed by using encoding:{}'.format(encoding), e)
            f.seek(0)
            continue
    return None


def marshal_object_for_ui(object):
    if isinstance(object, Enum):
        return object.value

    if isinstance(object, pd.Timestamp):
        return to_time_str(object)

    return object


def chrome_copy_header_to_dict(src):
    lines = src.split('\n')
    header = {}
    if lines:
        for line in lines:
            try:
                index = line.index(':')
                key = line[:index]
                value = line[index + 1:]
                if key and value:
                    header.setdefault(key.strip(), value.strip())
            except Exception:
                pass
    return header


def to_positive_number(number):
    if isinstance(number, numbers.Number):
        return abs(number)

    return 0


def multiple_number(number, factor):
    try:
        return number * factor
    except:
        return number


def add_to_map_list(the_map, key, value):
    result = []
    if key in the_map:
        result = the_map[key]
    else:
        the_map[key] = result

    if value not in result:
        result.append(value)


def iterate_with_step(data, sub_size=100):
    size = len(data)
    if size >= sub_size:
        step_count = int(size / sub_size)
        if size % sub_size:
            step_count = step_count + 1
    else:
        step_count = 1

    for step in range(step_count):
        if type(data) == pd.DataFrame or type(data) == pd.Series:
            yield data.iloc[sub_size * step:sub_size * (step + 1)]
        else:
            yield data[sub_size * step:sub_size * (step + 1)]


def url_unquote(url):
    return parse.unquote(url)


def parse_url_params(url):
    url = url_unquote(url)
    return parse.parse_qs(parse.urlsplit(url).query)


# the __all__ is generated
__all__ = ['first_item_to_float', 'second_item_to_float', 'add_func_to_value', 'to_float', 'pct_to_float', 'json_callback_param', 'fill_domain_from_dict', 'read_csv', 'marshal_object_for_ui', 'chrome_copy_header_to_dict', 'to_positive_number', 'multiple_number', 'add_to_map_list', 'iterate_with_step', 'url_unquote', 'parse_url_params']


# -*- coding: utf-8 -*-

from zvtm.contract import Exchange
from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import Recorder
from zvtm.domain import Stock
from zvtm.recorders.em import em_api


class EMStockRecorder(Recorder):
    provider = 'em'
    data_schema = Stock

    def run(self):
        for exchange in [Exchange.sh, Exchange.sz]:
            df = em_api.get_tradable_list(entity_type='stock', exchange=exchange)
            self.logger.info(df)
            df_to_db(df=df, data_schema=self.data_schema, provider=self.provider, force_update=True)


if __name__ == '__main__':
    recorder = EMStockRecorder()
    recorder.run()
# the __all__ is generated
__all__ = ['EMStockRecorder']


# -*- coding: utf-8 -*-

import time

import pandas as pd
import requests

from zvtm.api.kdata import generate_kdata_id
from zvtm.contract.recorder import FixedCycleDataRecorder
from zvtm.domain import Index, Index1dKdata
from zvtm.utils.time_utils import get_year_quarters, is_same_date


class ChinaIndexDayKdataRecorder(FixedCycleDataRecorder):
    entity_provider = 'exchange'
    entity_schema = Index

    provider = 'sina'
    data_schema = Index1dKdata
    url = 'http://vip.stock.finance.sina.com.cn/corp/go.php/vMS_MarketHistory/stockid/{}/type/S.phtml?year={}&jidu={}'

    def get_data_map(self):
        return {}

    def generate_domain_id(self, entity, original_data):
        return generate_kdata_id(entity.id, timestamp=original_data['timestamp'], level=self.level)

    def record(self, entity, start, end, size, timestamps):
        the_quarters = get_year_quarters(start)
        if not is_same_date(entity.timestamp, start) and len(the_quarters) > 1:
            the_quarters = the_quarters[1:]

        param = {
            'security_item': entity,
            'quarters': the_quarters,
            'level': self.level.value
        }

        security_item = param['security_item']
        quarters = param['quarters']
        level = param['level']

        result_df = pd.DataFrame()
        for year, quarter in quarters:
            query_url = self.url.format(security_item.code, year, quarter)
            response = requests.get(query_url)
            response.encoding = 'gbk'

            try:
                dfs = pd.read_html(response.text)
            except ValueError as error:
                self.logger.error(f'skip ({year}-{quarter:02d}){security_item.code}{security_item.name}({error})')
                time.sleep(10.0)
                continue

            if len(dfs) < 5:
                time.sleep(10.0)
                continue

            df = dfs[4].copy()
            df = df.iloc[1:]
            df.columns = ['timestamp', 'open', 'high', 'close', 'low', 'volume', 'turnover']
            df['name'] = security_item.name
            df['level'] = level
            df['timestamp'] = pd.to_datetime(df['timestamp'])
            df['provider'] = 'sina'

            result_df = pd.concat([result_df, df])

            self.logger.info(f'({security_item.code}{security_item.name})({year}-{quarter:02d})')
            time.sleep(10.0)

        result_df = result_df.sort_values(by='timestamp')

        return result_df.to_dict(orient='records')


__all__ = ['ChinaIndexDayKdataRecorder']

if __name__ == '__main__':
    ChinaIndexDayKdataRecorder().run()
# the __all__ is generated
__all__ = ['ChinaIndexDayKdataRecorder']


# -*- coding: utf-8 -*-

import io

import pandas as pd
import requests

from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import Recorder
from zvtm.domain import Stock, StockDetail
from zvtm.recorders.consts import DEFAULT_SH_HEADER, DEFAULT_SZ_HEADER
from zvtm.utils.time_utils import to_pd_timestamp


class ExchangeStockMetaRecorder(Recorder):
    data_schema = Stock
    provider = 'exchange'

    original_page_url = 'http://www.sse.com.cn/assortment/stock/list/share/'

    def run(self):
        url = 'http://query.sse.com.cn/security/stock/downloadStockListFile.do?csrcCode=&stockCode=&areaName=&stockType=1'
        resp = requests.get(url, headers=DEFAULT_SH_HEADER)
        self.download_stock_list(response=resp, exchange='sh')

        url = 'http://www.szse.cn/api/report/ShowReport?SHOWTYPE=xlsx&CATALOGID=1110&TABKEY=tab1&random=0.20932135244582617'
        resp = requests.get(url, headers=DEFAULT_SZ_HEADER)
        self.download_stock_list(response=resp, exchange='sz')

    def download_stock_list(self, response, exchange):
        df = None
        if exchange == 'sh':
            df = pd.read_csv(io.BytesIO(response.content), sep='\s+', encoding='GB2312', dtype=str,
                             parse_dates=['上市日期'], error_bad_lines=False)
            if df is not None:
                df = df.loc[:, ['公司代码', '公司简称', '上市日期']]

        elif exchange == 'sz':
            df = pd.read_excel(io.BytesIO(response.content), sheet_name='A股列表', dtype=str, parse_dates=['A股上市日期'])
            if df is not None:
                df = df.loc[:, ['A股代码', 'A股简称', 'A股上市日期']]

        if df is not None:
            df.columns = ['code', 'name', 'list_date']

            df = df.dropna(subset=['code'])

            # handle the dirty data
            # 600996,贵广网络,2016-12-26,2016-12-26,sh,stock,stock_sh_600996,,次新股,贵州,,
            df.loc[df['code'] == '600996', 'list_date'] = '2016-12-26'
            print(df[df['list_date'] == '-'])
            df['list_date'] = df['list_date'].apply(lambda x: to_pd_timestamp(x))
            df['exchange'] = exchange
            df['entity_type'] = 'stock'
            df['id'] = df[['entity_type', 'exchange', 'code']].apply(lambda x: '_'.join(x.astype(str)), axis=1)
            df['entity_id'] = df['id']
            df['timestamp'] = df['list_date']
            df = df.dropna(axis=0, how='any')
            df = df.drop_duplicates(subset=('id'), keep='last')
            df_to_db(df=df, data_schema=self.data_schema, provider=self.provider, force_update=False)
            # persist StockDetail too
            df_to_db(df=df, data_schema=StockDetail, provider=self.provider, force_update=False)
            self.logger.info(df.tail())
            self.logger.info("persist stock list successs")


__all__ = ['ExchangeStockMetaRecorder']

if __name__ == '__main__':
    recorder = ExchangeStockMetaRecorder()
    recorder.run()
# the __all__ is generated
__all__ = ['ExchangeStockMetaRecorder']


# -*- coding: utf-8 -*-
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule dividend_financing
from .dividend_financing import *
from .dividend_financing import __all__ as _dividend_financing_all
__all__ += _dividend_financing_all

# import all from submodule finance
from .finance import *
from .finance import __all__ as _finance_all
__all__ += _finance_all

# import all from submodule trading
from .trading import *
from .trading import __all__ as _trading_all
__all__ += _trading_all

# import all from submodule valuation
from .valuation import *
from .valuation import __all__ as _valuation_all
__all__ += _valuation_all


# -*- coding: utf-8 -*-
import logging
from enum import Enum

from zvtm.domain import IndexStock
from zvtm.tag.tag import StockTagger
# define tag
from zvtm.utils import pd_is_not_null
from zvtm.utils.time_utils import month_start_date, month_end_date

logger = logging.getLogger(__name__)


class StyleTag(Enum):
    # 成长股
    growth = 'growth'
    # 价值股
    value = 'value'


index_map_style = {
    # 国证成长（399370）
    # 在国证1000指数样本股中，选取主营业务收入增长率、净利润增长率和净资产收益率综合排名前332只
    'index_sz_399370': StyleTag.growth,
    # 国证价值（399371）
    # 在国证1000指数样本股中，选取每股收益与价格比率、每股经营现金流与价格比率、股息收益率、每股净资产与价格比率综合排名前332只
    'index_sz_399371': StyleTag.value
}


class StyleTagger(StockTagger):

    def tag(self, timestamp):
        for index_id in index_map_style:
            df = IndexStock.query_data(entity_id=index_id, start_timestamp=month_start_date(timestamp),
                                       end_timestamp=month_end_date(timestamp))
            if not pd_is_not_null(df):
                logger.error(f'no IndexStock data at {timestamp} for {index_id}')
                continue
            print(df)
            stock_tags = [self.get_tag_domain(entity_id=stock_id, timestamp=timestamp) for stock_id in
                          df['stock_id'].tolist()]
            for stock_tag in stock_tags:
                stock_tag.style_tag = index_map_style.get(index_id).value
            self.session.add_all(stock_tags)
            self.session.commit()


if __name__ == '__main__':
    StyleTagger().run()



# -*- coding: utf-8 -*-
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule jq_stock_meta_recorder
from .jq_stock_meta_recorder import *
from .jq_stock_meta_recorder import __all__ as _jq_stock_meta_recorder_all
__all__ += _jq_stock_meta_recorder_all

# import all from submodule jq_fund_meta_recorder
from .jq_fund_meta_recorder import *
from .jq_fund_meta_recorder import __all__ as _jq_fund_meta_recorder_all
__all__ += _jq_fund_meta_recorder_all

# import all from submodule jq_trade_day_recorder
from .jq_trade_day_recorder import *
from .jq_trade_day_recorder import __all__ as _jq_trade_day_recorder_all
__all__ += _jq_trade_day_recorder_all